import glob
import json
import os
import shutil
import sys
import unittest

import dpdata
import numpy as np
from monty.serialization import dumpfn, loadfn

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"

from pymatgen.io.vasp import Incar

from apex.core.calculator.calculator import make_calculator
from apex.core.common_equi import make_equi, post_equi

from .context import make_kspacing_kpoints, setUpModule


class TestEqui(unittest.TestCase):
    jdata = {
        "structures": ["confs/hp-Li"],
        "interaction": {
            "type": "vasp",
            "incar": "vasp_input/INCAR.rlx",
            "potcar_prefix": ".",
            "potcars": {"Li": "vasp_input/POTCAR"},
        },
        "relaxation": {
            "cal_type": "relaxation",
            "cal_setting": {"input_prop": "vasp_input/INCAR"},
        },
    }

    def tearDown(self):
        if os.path.exists("confs/hp-Li/relaxation"):
            shutil.rmtree("confs/hp-Li/relaxation")

    def test_make_equi(self):
        confs = self.jdata["structures"]
        inter_param = self.jdata["interaction"]
        relax_param = self.jdata["relaxation"]
        make_equi(confs, inter_param, relax_param)

        target_path = "confs/hp-Li/relaxation/relax_task"
        source_path = "vasp_input"

        incar0 = Incar.from_file(os.path.join("vasp_input", "INCAR"))
        incar1 = Incar.from_file(os.path.join(target_path, "INCAR"))
        self.assertTrue(incar0 == incar1)

        with open(os.path.join("vasp_input", "POTCAR")) as fp:
            pot0 = fp.read()
        with open(os.path.join(target_path, "POTCAR")) as fp:
            pot1 = fp.read()
        self.assertEqual(pot0, pot1)

        self.assertTrue(os.path.isfile(os.path.join(target_path, "KPOINTS")))

        task_json_file = os.path.join(target_path, "task.json")
        self.assertTrue(os.path.isfile(task_json_file))
        task_json = loadfn(task_json_file)
        self.assertEqual(task_json, relax_param)

        inter_json_file = os.path.join(target_path, "inter.json")
        self.assertTrue(os.path.isfile(inter_json_file))
        inter_json = loadfn(inter_json_file)
        self.assertEqual(inter_json, inter_param)

        self.assertTrue(os.path.islink(os.path.join(target_path, "POSCAR")))

    def test_post_equi(self):
        confs = self.jdata["structures"]
        inter_param = self.jdata["interaction"]
        relax_param = self.jdata["relaxation"]
        target_path = "confs/hp-Li/relaxation/relax_task"
        source_path = "equi/vasp"

        poscar = os.path.join(source_path, "POSCAR")
        make_equi(confs, inter_param, relax_param)
        shutil.copy(
            os.path.join(source_path, "OUTCAR"), os.path.join(target_path, "OUTCAR")
        )
        shutil.copy(
            os.path.join(source_path, "CONTCAR"), os.path.join(target_path, "CONTCAR")
        )
        post_equi(confs, inter_param)

        result_json_file = os.path.join(target_path, "result.json")
        result_json = loadfn(result_json_file)
        self.assertTrue(os.path.isfile(result_json_file))



import os
import shutil
import sys
import unittest

import numpy as np
from monty.serialization import loadfn

from apex.core.calculator.ABACUS import ABACUS
from apex.core.calculator.lib import abacus
from apex.core.calculator.lib import abacus_scf

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"


class TestABACUS(unittest.TestCase):
    def setUp(self):
        self.jdata = {
            "structures": ["confs/fcc-Al"],
            "interaction": {
                "type": "abacus",
                "incar": "abacus_input/INPUT",
                "potcar_prefix": "abacus_input",
                "potcars": {"Al": "Al_ONCV_PBE-1.0.upf"},
                "orb_files": {"Al": "Al_gga_9au_100Ry_4s4p1d.orb"},
            },
            "relaxation": {
                "cal_type": "relaxation",
                "cal_setting": {
                    "relax_pos": True,
                    "relax_shape": True,
                    "relax_vol": True,
                },
            },
        }

        self.conf_path = "confs/fcc-Al"
        self.equi_path = "confs/fcc-Al/relaxation/relax_task"
        self.source_path = "equi/abacus"
        if not os.path.exists(self.equi_path):
            os.makedirs(self.equi_path)

        self.confs = self.jdata["structures"]
        inter_param = self.jdata["interaction"]
        self.task_param = self.jdata["relaxation"]
        self.ABACUS = ABACUS(inter_param, os.path.join(self.conf_path, "STRU"))

    def tearDown(self):
        if os.path.exists("confs/fcc-Al/relaxation"):
            shutil.rmtree("confs/fcc-Al/relaxation")
        if os.path.isfile('apex_gen.log'):
            os.remove('apex_gen.log')

    def test_make_potential_files(self):
        if not os.path.exists(os.path.join(self.equi_path, "STRU")):
            with self.assertRaises(FileNotFoundError):
                self.ABACUS.make_potential_files(self.equi_path)
        shutil.copy(
            os.path.join(self.conf_path, "STRU"), os.path.join(self.equi_path, "STRU")
        )
        self.ABACUS.make_potential_files(self.equi_path)
        self.assertTrue(
            os.path.isfile(os.path.join(self.equi_path, "pp_orb/Al_ONCV_PBE-1.0.upf"))
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(self.equi_path, "pp_orb/Al_gga_9au_100Ry_4s4p1d.orb")
            )
        )
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "inter.json")))

    def test_make_input_file_1(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": True,
            "relax_shape": True,
            "relax_vol": False,
        }
        shutil.copy(
            os.path.join(self.conf_path, "STRU"), os.path.join(self.equi_path, "STRU")
        )
        self.ABACUS.make_input_file(self.equi_path, "relaxation", param)
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "task.json")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "KPT")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "INPUT")))
        abacus_input = abacus_scf.get_abacus_input_parameters(
            os.path.join(self.equi_path, "INPUT")
        )
        self.assertEqual(abacus_input["calculation"].lower(), "cell-relax")
        self.assertEqual(abacus_input["fixed_axes"].lower(), "volume")
        self.assertTrue(
            abacus.check_stru_fixed(os.path.join(self.equi_path, "STRU"), fixed=False)
        )

    def test_make_input_file_2(self):
        shutil.copy(
            os.path.join(self.conf_path, "STRU"), os.path.join(self.equi_path, "STRU")
        )
        self.ABACUS.make_input_file(self.equi_path, "relaxation", self.task_param)
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "task.json")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "KPT")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "INPUT")))
        abacus_input = abacus_scf.get_abacus_input_parameters(
            os.path.join(self.equi_path, "INPUT")
        )
        self.assertEqual(abacus_input["calculation"].lower(), "cell-relax")
        self.assertTrue(
            "fixed_axes" not in abacus_input or abacus_input["fixed_axes"] == "None"
        )
        self.assertTrue(
            abacus.check_stru_fixed(os.path.join(self.equi_path, "STRU"), fixed=False)
        )

    def test_make_input_file_3(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": True,
            "relax_shape": False,
            "relax_vol": False,
        }
        shutil.copy(
            os.path.join(self.conf_path, "STRU"), os.path.join(self.equi_path, "STRU")
        )
        self.ABACUS.make_input_file(self.equi_path, "relaxation", param)
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "task.json")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "KPT")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "INPUT")))
        abacus_input = abacus_scf.get_abacus_input_parameters(
            os.path.join(self.equi_path, "INPUT")
        )
        self.assertEqual(abacus_input["calculation"].lower(), "relax")
        self.assertTrue(
            "fixed_axes" not in abacus_input or abacus_input["fixed_axes"] == "None"
        )
        self.assertTrue(
            abacus.check_stru_fixed(os.path.join(self.equi_path, "STRU"), fixed=False)
        )

    def test_make_input_file_4(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": False,
            "relax_shape": True,
            "relax_vol": False,
        }
        shutil.copy(
            os.path.join(self.conf_path, "STRU"), os.path.join(self.equi_path, "STRU")
        )
        self.ABACUS.make_input_file(self.equi_path, "relaxation", param)
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "task.json")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "KPT")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "INPUT")))
        abacus_input = abacus_scf.get_abacus_input_parameters(
            os.path.join(self.equi_path, "INPUT")
        )
        self.assertEqual(abacus_input["calculation"].lower(), "cell-relax")
        self.assertEqual(abacus_input["fixed_axes"].lower(), "volume")
        self.assertTrue(
            abacus.check_stru_fixed(os.path.join(self.equi_path, "STRU"), fixed=True)
        )

    def test_make_input_file_5(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": False,
            "relax_shape": True,
            "relax_vol": True,
        }
        shutil.copy(
            os.path.join(self.conf_path, "STRU"), os.path.join(self.equi_path, "STRU")
        )
        self.ABACUS.make_input_file(self.equi_path, "relaxation", param)
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "task.json")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "KPT")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "INPUT")))
        abacus_input = abacus_scf.get_abacus_input_parameters(
            os.path.join(self.equi_path, "INPUT")
        )
        self.assertEqual(abacus_input["calculation"].lower(), "cell-relax")
        self.assertTrue(
            "fixed_axes" not in abacus_input or abacus_input["fixed_axes"] == "None"
        )
        self.assertTrue(
            abacus.check_stru_fixed(os.path.join(self.equi_path, "STRU"), fixed=True)
        )

    def test_make_input_file_kspacing(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": False,
            "relax_shape": True,
            "relax_vol": True,
            "kspacing": 0.1,
        }
        shutil.copy(
            os.path.join(self.conf_path, "STRU"), os.path.join(self.equi_path, "STRU")
        )
        self.ABACUS.make_input_file(self.equi_path, "relaxation", param)
        with open(os.path.join(self.equi_path, "KPT")) as f1:
            kpt = f1.read().strip().split("\n")[-1].split()
        self.assertEqual(kpt, ["9", "9", "9", "0", "0", "0"])

    def test_compuate(self):
        ret = self.ABACUS.compute(os.path.join(self.equi_path))
        self.assertIsNone(ret)
        shutil.copy(
            os.path.join(self.source_path, "INPUT"),
            os.path.join(self.equi_path, "INPUT"),
        )
        shutil.copy(
            os.path.join(self.conf_path, "STRU"), os.path.join(self.equi_path, "STRU")
        )
        os.mkdir(os.path.join(self.equi_path, "OUT.ABACUS"))
        shutil.copy(
            os.path.join(self.source_path, "running_cell-relax.log"),
            os.path.join(self.equi_path, "OUT.ABACUS", "running_cell-relax.log"),
        )
        ret = self.ABACUS.compute(self.equi_path)
        ret_ref = loadfn(os.path.join(self.source_path, "cell-relax.json"))

        def compare_dict(dict1, dict2):
            self.assertEqual(dict1.keys(), dict2.keys())
            for key in dict1:
                if type(dict1[key]) is dict:
                    compare_dict(dict1[key], dict2[key])
                else:
                    if type(dict1[key]) is np.ndarray:
                        np.testing.assert_almost_equal(
                            dict1[key], dict2[key], decimal=5
                        )
                    else:
                        self.assertTrue(dict1[key] == dict2[key])

        compare_dict(ret, ret_ref.as_dict())



import numpy as np
from typing import Dict, Union, Any, Tuple
from dflow.python import upload_packages
from apex.core.lib.trans_tools import (plane_miller_bravais_to_miller,
                                       direction_miller_bravais_to_miller)

upload_packages.append(__file__)


class SlabSlipSystem(object):
    __dict_atomic_system = {
        'fcc': {
            '001x100': {'plane': (0, 0, 1), 'x': (1, 0, 0), 'xy': (0, 1, 0),
                        'default_frac_slip_len': 1},
            '110x-110': {'plane': (1, 1, 0), 'x': (-1, 1, 0), 'xy': (0, 0, 1),
                         'default_frac_slip_len': float(np.sqrt(2))},
            '111x11-2': {'plane': (1, 1, 1), 'x': (1, 1, -2), 'xy': (-1, 1, 0),
                         'default_frac_slip_len': float(np.sqrt(6))},
            '111x-1-12': {'plane': (1, 1, 1), 'x': (-1, -1, 2), 'xy': (1, -1, 0),
                          'default_frac_slip_len': float(np.sqrt(6))},
            '111x-110': {'plane': (1, 1, 1), 'x': (-1, 1, 0), 'xy': (-1, -1, 2),
                         'default_frac_slip_len': float(np.sqrt(2))},
            '111x1-10': {'plane': (1, 1, 1), 'x': (1, -1, 0), 'xy': (1, 1, -2),
                         'default_frac_slip_len': float(np.sqrt(2))}
        },
        
        'bcc': {
            '001x100': {'plane': (0, 0, 1), 'x': (1, 0, 0), 'xy': (0, 1, 0),
                        'default_frac_slip_len': 1},
            '111x-110': {'plane': (1, 1, 1), 'x': (-1, 1, 0), 'xy': (-1, -1, 2),
                         'default_frac_slip_len': float(np.sqrt(2)/2)},
            '110x-111': {'plane': (1, 1, 0), 'x': (-1, 1, 1), 'xy': (0, 0, -1),
                         'default_frac_slip_len': float(np.sqrt(3)/2)},
            '110x1-1-1': {'plane': (1, 1, 0), 'x': (1, -1, -1), 'xy': (0, 0, 1),
                          'default_frac_slip_len': float(np.sqrt(3)/2)},
            '112x11-1': {'plane': (1, 1, 2), 'x': (1, 1, -1), 'xy': (-1, 1, 0),
                         'default_frac_slip_len': float(np.sqrt(3)/2)},
            '112x-1-11': {'plane': (1, 1, 2), 'x': (-1, -1, 1), 'xy': (1, -1, 0),
                          'default_frac_slip_len': float(np.sqrt(3)/2)},
            '123x11-1': {'plane': (1, 2, 3), 'x': (1, 1, -1), 'xy': (-2, 1, 0),
                         'default_frac_slip_len': float(np.sqrt(3)/2)},
            '123x-1-11': {'plane': (1, 2, 3), 'x': (-1, -1, 1), 'xy': (2, -1, 0),
                          'default_frac_slip_len': float(np.sqrt(3)/2)}
        },

        'hcp': {
            # Basal, cleavage and shear (non SF)
            '0001x2-1-10': {'plane': plane_miller_bravais_to_miller([0, 0, 0, 1]),
                            'x': direction_miller_bravais_to_miller([2, -1, -1, 0]),
                            'xy': direction_miller_bravais_to_miller([0, 1, -1, 0]),
                            'default_frac_slip_len': 1},

            # Basal, shear SF1 along x
            '0001x1-100': {'plane': plane_miller_bravais_to_miller([0, 0, 0, 1]),
                           'x': direction_miller_bravais_to_miller([1, -1, 0, 0]),
                           'xy': direction_miller_bravais_to_miller([0, 1, -1, 0]),
                           'default_frac_slip_len': float(np.sqrt(3))},

            # Basal, shear opposite to SF1 along x, climbing the hill
            '0001x10-10': {'plane': plane_miller_bravais_to_miller([0, 0, 0, 1]),
                           'x': direction_miller_bravais_to_miller([1, 0, -1, 0]),
                           'xy': direction_miller_bravais_to_miller([0, 1, -1, 0]),
                           'default_frac_slip_len': float(np.sqrt(3))},

            # Prism I, cleavage and SF1 along x
            '01-10x-2110': {'plane': plane_miller_bravais_to_miller([0, 1, -1, 0]),
                            'x': direction_miller_bravais_to_miller([-2, 1, 1, 0]),
                            'xy': direction_miller_bravais_to_miller([0, 0, 0, -1]),
                            'default_frac_slip_len': 1},

            # Prism I, cleavage and SF2 along xy
            '01-10x0001': {'plane': plane_miller_bravais_to_miller([0, 1, -1, 0]),
                            'x': direction_miller_bravais_to_miller([0, 0, 0, 1]),
                            'xy': direction_miller_bravais_to_miller([-2, 1, 1, 0]),
                            'default_frac_slip_len': (0, 0, 1)},

            # Prism I, shear SF2 along x
            '01-10x-2113': {'plane': plane_miller_bravais_to_miller([0, 1, -1, 0]),
                            'x': direction_miller_bravais_to_miller([-2, 1, 1, 3]),
                            'xy': direction_miller_bravais_to_miller([0, 0, 0, -1]),
                            'default_frac_slip_len': (1, 0, 1)},

            # Prism II, cleavage
            '-12-10x-1010': {'plane': plane_miller_bravais_to_miller([-1, 2, -1, 0]),
                             'x': direction_miller_bravais_to_miller([-1, 0, 1, 0]),
                             'xy': direction_miller_bravais_to_miller([0, 0, 0, -1]),
                             'default_frac_slip_len': float(np.sqrt(3))},

            # Prism II, cleavage along c
            '-12-10x0001': {'plane': plane_miller_bravais_to_miller([-1, 2, -1, 0]),
                            'x': direction_miller_bravais_to_miller([0, 0, 0, 1]),
                            'xy': direction_miller_bravais_to_miller([-1, 0, 1, 0]),
                            'default_frac_slip_len': (0, 0, 1)},

            # Pyramidal I, cleavage and <a> slip along x
            '01-11x-2110': {'plane': plane_miller_bravais_to_miller([0, 1, -1, 1]),
                            'x': direction_miller_bravais_to_miller([-2, 1, 1, 0]),
                            'xy': direction_miller_bravais_to_miller([-1, 2, -1, -3]),
                            'default_frac_slip_len': 1},

            # Pyramidal I, <c+a> slip along x
            '01-11x-12-1-3': {'plane': plane_miller_bravais_to_miller([0, 1, -1, 1]),
                            'x': direction_miller_bravais_to_miller([-1, 2, -1, -3]),
                            'xy': direction_miller_bravais_to_miller([2, -1, -1, 0]),
                            'default_frac_slip_len': (1, 0, 1)},

            # Pyramidal I, shear SF2 along x
            '01-11x0-112': {'plane': plane_miller_bravais_to_miller([0, 1, -1, 1]),
                            'x': direction_miller_bravais_to_miller([0, -1, 1, 2]),
                            'xy': direction_miller_bravais_to_miller([-1, 2, -1, -3]),
                            'default_frac_slip_len': (float(np.sqrt(3)), 0, 2)},

            # Pyramidal II, Cleavage
            '-12-12x10-10': {'plane': plane_miller_bravais_to_miller([-1, 2, -1, 2]),
                             'x': direction_miller_bravais_to_miller([1, 0, -1, 0]),
                             'xy': direction_miller_bravais_to_miller([1, -2, 1, 3]),
                             'default_frac_slip_len': float(np.sqrt(3))},

            # Pyramidal II, shear SF2 along x
            '-12-12x1-213': {'plane': plane_miller_bravais_to_miller([-1, 2, -1, 2]),
                             'x': direction_miller_bravais_to_miller([1, -2, 1, 3]),
                             'xy': direction_miller_bravais_to_miller([-1, 0, 1, 0]),
                             'default_frac_slip_len': (1, 0, 1)}
        }
    }

    @classmethod
    def atomic_system_dict(cls):
        return cls.__dict_atomic_system

    @classmethod
    def hint_string(cls):
        print_str = 'structure  \tplane_index    \tslip_direction\n'
        for struct in cls.__dict_atomic_system.keys():
            for orient in cls.__dict_atomic_system[struct].keys():
                plane, slip = orient.split('x')
                print_str += f'{struct}       \t{plane}         \t{slip}\n'
        return print_str


if __name__ == '__main__':
    print(SlabSlipSystem.hint_string())



import os, glob, pathlib, shutil, subprocess
from pathlib import Path
from typing import List
from dflow.python import (
    OP,
    OPIO,
    OPIOSign,
    Artifact,
    upload_packages
)
try:
    from monty.serialization import loadfn
    from apex.utils import return_prop_list
    from apex.core.common_prop import make_property_instance
    from .utils import recursive_search
except:
    pass

upload_packages.append(__file__)


class DistributeProps(OP):
    """
    OP class for distribution
    of individual property test steps
    """
    # TODO: add API for complex property test superOP
    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            "input_work_path": Artifact(Path),
            "param": dict
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            "orig_work_path": Artifact(List[Path]),
            "flow_id": List[str],
            "path_to_prop": List[str],
            "prop_param": List[dict],
            "inter_param": List[dict],
            "do_refine": List[bool],
            "nflows": int
        })

    @OP.exec_sign_check
    def execute(
            self,
            op_in: OPIO,
    ) -> OPIO:
        input_work_path = op_in["input_work_path"]
        param = op_in["param"]

        cwd = Path.cwd()
        os.chdir(input_work_path)
        confs = param["structures"]
        interaction = param["interaction"]
        properties = param["properties"]

        conf_dirs = []
        flow_id_list = []
        path_to_prop_list = []
        prop_param_list = []
        do_refine_list = []
        for conf in confs:
            conf_dirs.extend(glob.glob(conf))
        conf_dirs.sort()
        for ii in conf_dirs:
            for jj in properties:
                if jj.get('skip', False):
                    continue
                if 'init_from_suffix' and 'output_suffix' in jj:
                    do_refine = True
                    suffix = jj['output_suffix']
                elif 'reproduce' in jj and jj['reproduce']:
                    do_refine = False
                    suffix = 'reprod'
                elif 'suffix' in jj and jj['suffix']:
                    do_refine = False
                    suffix = str(jj['suffix'])
                else:
                    do_refine = False
                    suffix = '00'

                property_type = jj["type"]
                path_to_prop_list.append(os.path.join(ii, property_type + "_" + suffix))
                prop_param_list.append(jj)
                do_refine_list.append(do_refine)
                flow_id_list.append(ii + '-' + property_type + '-' + suffix)

        nflow = len(path_to_prop_list)
        orig_work_path_list = [input_work_path] * nflow
        inter_param_list = [interaction] * nflow

        op_out = OPIO({
            "orig_work_path": orig_work_path_list,
            "flow_id": flow_id_list,
            "path_to_prop": path_to_prop_list,
            "prop_param": prop_param_list,
            "inter_param": inter_param_list,
            "do_refine": do_refine_list,
            "nflows": nflow
        })
        return op_out


class PropsMake(OP):
    """
    OP class for making calculation tasks (make property)
    """
    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            'input_work_path': Artifact(Path),
            'path_to_prop': str,
            'prop_param': dict,
            'inter_param': dict,
            'do_refine': bool
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            'output_work_path': Artifact(Path),
            'task_names': List[str],
            'njobs': int,
            'task_paths': Artifact(List[Path])
        })

    @OP.exec_sign_check
    def execute(
            self,
            op_in: OPIO,
    ) -> OPIO:
        from apex.core.common_prop import make_property_instance
        from apex.core.calculator.calculator import make_calculator

        input_work_path = op_in["input_work_path"]
        path_to_prop = op_in["path_to_prop"]
        prop_param = op_in["prop_param"]
        inter_param = op_in["inter_param"]
        do_refine = op_in["do_refine"]

        cwd = Path.cwd()
        os.chdir(input_work_path)
        abs_path_to_prop = input_work_path / path_to_prop
        conf_path = abs_path_to_prop.parent
        prop_name = abs_path_to_prop.name
        path_to_equi = conf_path / "relaxation" / "relax_task"
        prop = make_property_instance(prop_param, inter_param)
        task_list = prop.make_confs(abs_path_to_prop, path_to_equi, do_refine)
        for kk in task_list:
            poscar = os.path.join(kk, "POSCAR")
            inter = make_calculator(inter_param, poscar)
            inter.make_potential_files(kk)
            # dlog.debug(prop.task_type())  ### debug
            inter.make_input_file(kk, prop.task_type(), prop.task_param())
        prop.post_process(
            task_list
        )  # generate same KPOINTS file for elastic when doing VASP

        task_list.sort()
        os.chdir(input_work_path)
        task_list_str = glob.glob(path_to_prop + '/' + 'task.*')

        all_jobs = task_list
        njobs = len(all_jobs)
        jobs = []
        for job in all_jobs:
            jobs.append(pathlib.Path(job))

        os.chdir(cwd)
        op_out = OPIO({
            "output_work_path": input_work_path,
            "task_names": task_list_str,
            "njobs": njobs,
            "task_paths": jobs
        })
        return op_out


class PropsPost(OP):
    """
    OP class for analyzing calculation results (post property)
    """

    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            'input_post': Artifact(Path, sub_path=False),
            'input_all': Artifact(Path),
            'prop_param': dict,
            'inter_param': dict,
            'task_names': List[str],
            'path_to_prop': str
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            'output_post': Artifact(Path)
        })

    @OP.exec_sign_check
    def execute(self, op_in: OPIO) -> OPIO:
        from apex.core.common_prop import make_property_instance
        cwd = os.getcwd()
        input_post = op_in["input_post"]
        input_all = op_in["input_all"]
        prop_param = op_in["prop_param"]
        inter_param = op_in["inter_param"]
        task_names = op_in["task_names"]
        path_to_prop = op_in["path_to_prop"]
        calculator = inter_param["type"]
        copy_dir_list_input = [path_to_prop.split('/')[0]]
        os.chdir(input_all)
        copy_dir_list = []
        for ii in copy_dir_list_input:
            copy_dir_list.extend(glob.glob(ii))

        # find path of finished tasks
        os.chdir(op_in['input_post'])
        src_path = recursive_search(copy_dir_list)
        if not src_path:
            raise RuntimeError(f'Fail to find input work path after slices!')

        if calculator in ['vasp', 'abacus']:
            os.chdir(input_post)
            for ii in task_names:
                shutil.copytree(os.path.join(ii, "backward_dir"), ii, dirs_exist_ok=True)
                shutil.rmtree(os.path.join(ii, "backward_dir"))
            os.chdir(input_all)
            shutil.copytree(input_post, './', dirs_exist_ok=True)
        else:
            os.chdir(input_all)
            #src_path = str(input_post) + str(local_path)
            shutil.copytree(src_path, './', dirs_exist_ok=True)

        if ("cal_setting" in prop_param
                and "overwrite_interaction" in prop_param["cal_setting"]):
            inter_param = prop_param["cal_setting"]["overwrite_interaction"]

        abs_path_to_prop = Path.cwd() / path_to_prop

        prop = make_property_instance(prop_param, inter_param)
        prop.compute(
            os.path.join(abs_path_to_prop, "result.json"),
            os.path.join(abs_path_to_prop, "result.out"),
            abs_path_to_prop,
        )
        # remove potential files in each md task
        os.chdir(abs_path_to_prop)
        cmd = "for kk in task.*; do cd $kk; rm *.pb; cd ..; done"
        subprocess.call(cmd, shell=True)

        os.chdir(cwd)
        out_path = Path(cwd) / 'retrieve_pool'
        os.mkdir(out_path)
        shutil.copytree(input_all / path_to_prop,
                        out_path / path_to_prop, dirs_exist_ok=True)

        op_out = OPIO({
            'output_post': abs_path_to_prop
        })
        return op_out


class CollectProps(OP):
    """
    OP class for collect property tasks
    """

    def __init__(self):
        pass

    @classmethod
    def get_input_sign(cls):
        return OPIOSign({
            'input_post': Artifact(Path, sub_path=False),
            'input_all': Artifact(Path),
            'param': dict
        })

    @classmethod
    def get_output_sign(cls):
        return OPIOSign({
            'retrieve_path': Artifact(List[Path])
        })

    @OP.exec_sign_check
    def execute(self, op_in: OPIO) -> OPIO:
        cwd = os.getcwd()
        input_post = op_in["input_post"]
        input_all = op_in["input_all"]
        param = op_in["param"]
        confs = param["structures"]
        copy_dir_list_input = [conf.split('/')[0] for conf in confs]
        os.chdir(op_in['input_all'])
        copy_dir_list = []
        for ii in copy_dir_list_input:
            copy_dir_list.extend(glob.glob(ii))
        os.chdir(input_post)

        src_path = recursive_search(copy_dir_list)
        if not src_path:
            raise RuntimeError(f'Fail to find input work path after slices!')
        shutil.copytree(src_path, input_all, dirs_exist_ok=True)

        for ii in copy_dir_list:
            shutil.copytree(input_all / ii, ii, dirs_exist_ok=True)

        retrieve_path = [Path(ii) for ii in copy_dir_list]

        op_out = OPIO({
            'retrieve_path': retrieve_path
        })
        return op_out




#!/usr/bin/env python3

from __future__ import division

import argparse
import os
import sys

import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate as INT
from scipy.interpolate import *
from scipy.misc import derivative
from scipy.optimize import curve_fit, fsolve, leastsq, minimize, root
from dflow.python import upload_packages
upload_packages.append(__file__)

kb = 1.3806488e-23  # J K^-1
kb_ev = 8.6173324e-05  # eV K^-1
h = 6.62606957e-34  # J.s
h_ev = 4.135667516e-15  # eV s
hb = 1.054571726e-34  # J.s
hb_ev = 6.58211928e-16  # eV s
mu = 1.660538921e-27  # kg
me = 9.1093821545e-31  # kg
NA = 6.02214129e23  # number per mol

eV2GPa = 1.602176565e2
eV2mol = 9.648455461e4  # eV/K --> J/mol/K


def __version__():
    return "1.2.5"


def get_eos_list_4p():
    eos_list_4p = [
        "murnaghan",
        "birch",
        "BM4",
        "mBM4",
        "mBM4poly",
        "rBM4",
        "rPT4",
        "LOG4",
        "vinet",
        "Li4p",
        "universal",
        "morse",
        "morse_AB",
        "mie",
        "mie_simple",
        "SJX_v2",
    ]
    return eos_list_4p


def get_eos_list_5p():
    eos_list_5p = ["BM5", "mBM5", "rBM5", "mBM5poly", "rPT5", "LOG5", "TEOS", "SJX_5p"]
    return eos_list_5p


def get_eos_list_6p():
    eos_list_6p = ["morse_6p"]
    return eos_list_6p


def get_eos_list_3p():
    eos_list_3p = ["morse_3p"]
    return eos_list_3p


def get_eos_list():
    LIST_ALL = (
        get_eos_list_3p() + get_eos_list_4p() + get_eos_list_5p() + get_eos_list_6p()
    )
    return LIST_ALL


# ----------------------------------------------------------------------------------------
def res_murnaghan(pars, y, x):
    return y - murnaghan(x, pars)


def murnaghan(vol, pars):
    """
    Four-parameters murnaghan EOS.
    From PRB 28,5480 (1983)
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    xx = (v0 / vol) ** bp
    A = e0 - b0 * v0 / (bp - 1)
    B = b0 * vol / bp
    ee = A + B * (1 + xx / (bp - 1))

    # ee = e0+b0*vol/bp*(((v0/vol)**bp)/(bp-1)+1) -b0*v0/(bp-1)

    return ee


# ----------------------------------------------------------------------------------------


def res_birch(pars, y, x):
    return y - birch(x, pars)


def birch(v, parameters):
    """
    From Intermetallic compounds: Principles and Practice, Vol. I: Princples
    Chapter 9 pages 195-210 by M. Mehl. B. Klein, D. Papaconstantopoulos
    paper downloaded from Web

    case where n=0
    """
    e0 = parameters[0]
    b0 = parameters[1]
    bp = parameters[2]
    v0 = parameters[3]

    e = (
        e0
        + 9.0 / 8.0 * b0 * v0 * ((v0 / v) ** (2.0 / 3.0) - 1.0) ** 2
        + 9.0 / 16.0 * b0 * v0 * (bp - 4.0) * ((v0 / v) ** (2.0 / 3.0) - 1.0) ** 3
    )

    return e


# ----------------------------------------------------------------------------------------
def res_mBM4(pars, y, x):
    return y - mBM4(x, pars)


def calc_props_mBM4(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    bpp = (-74 + 45 * bp - 9 * bp * bp) / (9 * b0)
    props = [e0, b0, bp, v0, bpp]
    return props


def mBM4(vol, pars):
    """
    Birch-Murnaghan 4 pars equation from PRB 70, 224107, 3-order BM
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    a = e0 + 9 * b0 * v0 * (4 - bp) / 2.0
    b = -9 * b0 * v0 ** (4.0 / 3) * (11 - 3 * bp) / 2.0
    c = 9 * b0 * v0 ** (5.0 / 3) * (10 - 3 * bp) / 2.0
    d = -9 * b0 * v0**2 * (3 - bp) / 2.0

    n = 1  # 1 as mBM, 2 as BM
    VV = np.power(vol, -n / 3)
    ee = a + b * VV + c * VV**2 + d * VV**3

    return ee


def res_mBM5(pars, y, x):
    return y - mBM5(x, pars)


def mBM5(vol, pars):
    """
    modified BM5 EOS, Shang SL comput mater sci, 2010: 1040-1048
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    b2p = pars[4]

    """
    # copy from ShunLi's matlab scripts.
    a = (8 * e0 + 3 * b0 * (122 + 9 * b0 * b2p - 57 * bp + 9 * bp * bp) * v0) / 8
    b = (-3 * b0 * (107 + 9 * b0 * b2p - 54 * bp + 9 * bp * bp) * v0 ** (4 / 3)) / 2
    c = (9 * b0 * (94 + 9 * b0 * b2p - 51 * bp + 9 * bp * bp) * v0 ** (5 / 3)) / 4
    d = (-3 * b0 * (83 + 9 * b0 * b2p - 48 * bp + 9 * bp * bp) * v0 ** 2) / 2
    e = (3 * b0 * (74 + 9 * b0 * b2p - 45 * bp + 9 * bp * bp) * v0 ** (7 / 3)) / 8
    """
    # retype according to formula in the article
    a = e0 + 3 * b0 * v0 * (122 + 9 * b0 * b2p - 57 * bp + 9 * bp * bp) / 8
    b = -3 * b0 * v0 ** (4 / 3) * (107 + 9 * b0 * b2p - 54 * bp + 9 * bp * bp) / 2
    c = 9 * b0 * v0 ** (5 / 3) * (94 + 9 * b0 * b2p - 51 * bp + 9 * bp * bp) / 4
    d = -3 * b0 * v0**2 * (83 + 9 * b0 * b2p - 48 * bp + 9 * bp * bp) / 2
    e = 3 * b0 * v0 ** (7 / 3) * (74 + 9 * b0 * b2p - 45 * bp + 9 * bp * bp) / 8

    VV = np.power(vol, -1 / 3)
    ee = a + b * VV + c * VV**2 + d * VV**3 + e * VV**4

    return ee


# ----------------------------------------------------------------------------------------
def res_mBM4poly(pars, y, x):
    return y - mBM4poly(x, pars)


def mBM4poly(vol, parameters):
    """
    modified BM5 EOS, Shang SL comput mater sci, 2010: 1040-1048, original expressions.
    """
    a = parameters[0]
    b = parameters[1]
    c = parameters[2]
    d = parameters[3]
    e = 0

    n = 1  # 1 as mBM, 2 as BM
    VV = np.power(vol, -1 / 3)
    E = a + b * VV + c * VV**2 + d * VV**3 + e * VV**4
    return E


def calc_v0_mBM4poly(x, pars):
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = 0

    f = (
        (4 * e) / (3 * x ** (7 / 3))
        + d / x**2
        + (2 * c) / (3 * x ** (5 / 3))
        + b / (3 * x ** (4 / 3))
    ) * eV2GPa
    return f


def calc_props_mBM4poly(pars):
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = 0

    v0 = (
        4 * c**3
        - 9 * b * c * d
        + np.sqrt((c**2 - 3 * b * d) * (4 * c**2 - 3 * b * d) ** 2)
    )
    v0 = -v0 / b**3
    b0 = (
        (28 * e) / (9 * v0 ** (10 / 3))
        + (2 * d) / v0**3
        + (10 * c) / (9 * v0 ** (8 / 3))
        + (4 * b) / (9 * v0 ** (7 / 3))
    ) * v0
    bp = (98 * e + 54 * d * v0 ** (1 / 3) + 25 * c * v0 ** (2 / 3) + 8 * b * v0) / (
        42 * e + 27 * d * v0 ** (1 / 3) + 15 * c * v0 ** (2 / 3) + 6 * b * v0
    )
    b2p = (
        v0 ** (8 / 3)
        * (
            9 * d * (14 * e + 5 * c * v0 ** (2 / 3) + 8 * b * v0)
            + 2
            * v0 ** (1 / 3)
            * (126 * b * e * v0 ** (1 / 3) + 5 * c * (28 * e + b * v0))
        )
    ) / (2 * (14 * e + 9 * d * v0 ** (1 / 3) + 5 * c * v0 ** (2 / 3) + 2 * b * v0) ** 3)
    e0 = mBM4poly(v0, pars)

    props = [e0, b0, bp, v0, b2p]
    return props


# ---------------------------------


def res_mBM5poly(pars, y, x):
    return y - mBM5poly(x, pars)


def mBM5poly(vol, pars):
    """
    modified BM5 EOS, Shang SL comput mater sci, 2010: 1040-1048, original expressions.
    """
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = pars[4]

    n = 1  # 1 as mBM, 2 as BM
    VV = np.power(vol, -1 / 3)
    E = a + b * VV + c * VV**2 + d * VV**3 + e * VV**4
    return E


def calc_v0_mBM5poly(x, pars):
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = pars[4]

    f = (
        (4 * e) / (3 * x ** (7 / 3))
        + d / x**2
        + (2 * c) / (3 * x ** (5 / 3))
        + b / (3 * x ** (4 / 3))
    ) * eV2GPa
    return f


def calc_props_mBM5poly(pars):
    a = pars[0]
    b = pars[1]
    c = pars[2]
    d = pars[3]
    e = pars[4]
    n = 1

    # guess v0
    vtest = np.linspace(1, 100, 101)
    etest = mBM5poly(vtest, pars)
    vg = vtest[etest.argmin()]

    # sol = root(calc_v0_mBM5poly, vg, args=pars)
    # v0 = sol.x[0]
    sol = fsolve(calc_v0_mBM5poly, vg, args=pars)
    v0 = sol

    b0 = (
        (28 * e) / (9 * v0 ** (10 / 3))
        + (2 * d) / v0**3
        + (10 * c) / (9 * v0 ** (8 / 3))
        + (4 * b) / (9 * v0 ** (7 / 3))
    ) * v0
    bp = (98 * e + 54 * d * v0 ** (1 / 3) + 25 * c * v0 ** (2 / 3) + 8 * b * v0) / (
        42 * e + 27 * d * v0 ** (1 / 3) + 15 * c * v0 ** (2 / 3) + 6 * b * v0
    )
    b2p = (
        v0 ** (8 / 3)
        * (
            9 * d * (14 * e + 5 * c * v0 ** (2 / 3) + 8 * b * v0)
            + 2
            * v0 ** (1 / 3)
            * (126 * b * e * v0 ** (1 / 3) + 5 * c * (28 * e + b * v0))
        )
    ) / (2 * (14 * e + 9 * d * v0 ** (1 / 3) + 5 * c * v0 ** (2 / 3) + 2 * b * v0) ** 3)

    """
    dEdV = -b / 3 * v0**(-4 / 3) - 2 / 3 * c * \
        v0**(-5 / 3) - d * v0**(-2) - 4 / 3 * e * v0**(-7 / 3)
    dEdV2 = 4 / 9 * b * v0**(-7 / 3) + 10 / 9 * c * \
        v0**(-8 / 3) + 2 * d * v0**(-3) + 28 / 9 * e * v0**(-10 / 3)
    dEdV3 = -28 / 27 * b * \
        v0**(-10 / 3) - 80 / 27 * c * v0**(-11 / 3) - 6 * \
        d * v0**(-4) - 280 / 27 * e * v0**(-13 / 3)
    dEdV4 = 280 / 81 * b * v0**(-13 / 3) + 880 / 81 * c * \
        v0**(-14 / 3) + 24 * d * v0**(-5) + 3640 / 81 * e * v0**(-16 / 3)

    P = -v0 * dEdV
    dBdV = dEdV2 + v0 * dEdV3
    dPdV = -dEdV - v0 * dEdV2
    dBdV2 = dEdV3 + dEdV3 + v0 * dEdV4
    dPdV2 = -dEdV2 - dEdV2 - v0 * dEdV3

    b0 = v0 * dEdV2
    bp = dBdV / dPdV
    b2p = (dBdV2 * dPdV - dPdV2 * dBdV) / dPdV**3
    """
    e0 = mBM5poly(v0, pars)
    props = [e0, b0, bp, v0, b2p]
    return props


# ----------------------------------------------------------------------------------------


def res_BM4(pars, y, x):
    return y - BM4(x, pars)


def calc_props_BM4(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = (-143 + 63 * bp - 9 * bp * bp) / (9 * b0)
    props = [e0, b0, bp, v0, bpp]
    return props


def BM4(vol, pars):
    """
    Birch-Murnaghan 4 pars equation from PRB 70, 224107, 3-order
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    eta = (v0 / vol) ** (1.0 / 3.0)
    e = e0 + 9.0 * b0 * v0 / 16 * (eta**2 - 1) ** 2 * (
        6 + bp * (eta**2 - 1.0) - 4.0 * eta**2
    )
    return e


# ----------------------------------------------------------------------------------------
def res_BM5(pars, y, x):
    return y - BM5(x, pars)


def BM5(vol, pars):
    """
    Birch-Murnaghan 5 pars equation from PRB 70, 224107, 4-Order
    """
    e0 = pars[0]
    b0 = pars[1]
    b0p = pars[2]
    v0 = pars[3]
    b0pp = pars[4]

    t1 = (v0 / vol) ** (1.0 / 3.0)
    t2 = t1**2
    t3 = t2 - 1.0
    t4 = t3**2 / 4.0
    t5 = b0p**2

    ee = e0 + 3.0 / 8.0 * b0 * v0 * t4 * (
        9.0 * t4 * b0 * b0pp
        + 9.0 * t4 * t5
        - 63.0 * t4 * b0p
        + 143.0 * t4
        + 6.0 * b0p * t3
        - 24.0 * t2
        + 36.0
    )

    return ee


def rBM4(vol, pars):
    """
    Implementions as Alberto Otero-de-la-Roza, i.e. rBM4 is used here
    Comput Physics Comm, 2011, 182: 1708-1720
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    x = v0 / vol
    f = 0.5 * (x ** (2.0 / 3) - 1)
    E = e0 + 4.5 * v0 * b0 * f**2 * (1 + (bp - 4) * f)
    return E


def res_rBM4(pars, y, x):
    res = y - rBM4(x, pars)
    return res


def rBM4_pv(vol, pars):
    """
    Implementions as Alberto Otero-de-la-Roza, i.e. rBM4 is used here
    Comput Physics Comm, 2011, 182: 1708-1720
    Fit for V-P relations
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    x = v0 / vol
    f = 0.5 * (x ** (2.0 / 3) - 1)
    P = 1.5 * b0 * (2 * f + 1) ** (2.5) * (2 + 3 * (bp - 4) * f)
    return P


def res_rBM4_pv(par, y, x):
    res = y - rBM4_pv(x, pars)
    return res


def rBM5(vol, pars):
    """
    Implementions as Alberto Otero-de-la-Roza, i.e. rBM5 is used here
    Comput Physics Comm, 2011, 182: 1708-1720
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = pars[4]

    x = v0 / vol
    f = 0.5 * (x ** (2.0 / 3) - 1)
    H = b0 * bpp + b0 * b0

    E = e0 + 3 / 8 * v0 * b0 * f**2 * (
        (9 * H - 63 * bp + 143) * f**2 + 12 * (bp - 4) * f + 12
    )
    return E


def res_rBM5(pars, y, x):
    res = y - rBM5(x, pars)
    return res


def rBM5_pv(vol, pars):
    """
    Implementions as Alberto Otero-de-la-Roza, i.e. rBM5 is used here
    Comput Physics Comm, 2011, 182: 1708-1720
    Fit for V-P relations
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = pars[4]

    x = v0 / vol
    f = 0.5 * (x ** (2.0 / 3) - 1)
    H = b0 * bpp + b0 * b0

    P = (
        0.5
        * b0
        * (2 * f + 1) ** (2.5)
        * ((9 * H - 63 * bp + 143) * f**2 + 9 * (bp - 4) * f + 6)
    )
    return P


def res_rBM5_pv(par, y, x):
    res = y - rBM5_pv(x, pars)
    return res


# ----------------------------------------------------------------------------------------


def res_universal(pars, y, x):
    return y - universal(x, pars)


def universal(vol, parameters):
    """
    Universal equation of state(Vinet P et al., J. Phys.: Condens. Matter 1, p1941 (1989))
    """
    e0 = parameters[0]
    b0 = parameters[1]
    bp = parameters[2]
    v0 = parameters[3]

    t1 = b0 * v0
    t2 = bp - 1.0
    t3 = (vol / v0) ** (1.0 / 3.0)
    t4 = np.exp(-3.0 / 2.0 * t2 * (-1.0 + t3))
    t5 = t2**2
    t6 = 1.0 / t5
    e = (
        e0
        - 2.0 * t1 * t4 * (3.0 * t3 * bp - 3.0 * t3 + 5.0 - 3.0 * bp) * t6
        + 4.0 * t1 * t6
    )

    return e


# ----------------------------------------------------------------------------------------
def res_LOG4(pars, y, x):
    return y - LOG4(x, pars)


def LOG4(vol, pars):
    """
    Natrual strain (Poirier-Tarantola)EOS with 4 paramters
    Seems only work in near-equillibrium range.
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    t1 = b0 * v0
    t2 = np.log(v0 / vol)
    t3 = t2**2
    t4 = t3 * t2
    ee = e0 + t1 * t3 / 2.0 + t1 * t4 * bp / 6.0 - t1 * t4 / 3.0
    """
    # write follows ShunLi's
    xx = np.log(v0)
    a = e0 + b0 * v0 * (3 * xx**2 + (bp - 2) * xx**3) / 6
    b = -b0 * v0 * (2 * xx + (bp - 2) * xx**2) / 2
    c = b0 * v0 * (1 + (bp - 2) * xx) / 2
    d = -b0 * v0 * (bp - 2) / 6
    VV = np.log(vol)
    ee = a + b * VV + c * VV**2 + d * VV**3
    """
    return ee


def calc_props_LOG4(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    bpp = (-3 + 3 * bp - bp * bp) / b0
    props = [e0, b0, bp, v0, bpp]
    return props


# ----------------------------------------------------------------------------------------


def rPT4(vol, pars):
    """
    Natrual strain EOS with 4 paramters
    Seems only work in near-equillibrium range.
    Implementions as Alberto Otero-de-la-Roza, i.e. rPT4 is used here
    Comput Physics Comm, 2011, 182: 1708-1720,
    in their article, labeled as PT3 (3-order), however, we mention it as
    rPT4 for 4-parameters EOS.
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    x = vol / v0
    fn = 1 / 3 * np.log(x)
    E = e0 + 4.5 * b0 * v0 * fn**2 * (-(bp - 2) * fn + 1)
    return E


def res_rPT4(pars, y, x):
    res = y - rPT4(x, pars)
    return res


def rPT4_pv(vol, pars):
    """
    Natrual strain (Poirier-Tarantola)EOS with 4 paramters
    Seems only work in near-equillibrium range.
    Implementions as Alberto Otero-de-la-Roza, i.e. rPT4 is used here
    Comput Physics Comm, 2011, 182: 1708-1720,
    in their article, labeled as PT3 (3-order), however, we mention it as
    rPT4 for 4-parameters EOS.
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    x = vol / v0
    fn = 1 / 3 * np.log(x)
    P = -1.5 * b0 * fn * np.exp(-3 * fn) * (-3 * (bp - 2) * fn + 1)
    return P


def res_rPT4_pv(pars, y, x):
    res = y - rPT4_pv(x, pars)
    return res


# ----------------------------------------------------------------------------------------
def res_LOG5(pars, y, x):
    return y - LOG5(x, pars)


def LOG5(vol, parameters):
    """
    Natrual strain (Poirier-Tarantola)EOS with 5 paramters
    """
    e0 = parameters[0]
    b0 = parameters[1]
    b0p = parameters[2]
    v0 = parameters[3]
    b0pp = parameters[4]

    t1 = b0 * v0
    t2 = np.log(v0 / vol)
    t3 = t2**2
    t4 = t3**2
    t5 = b0**2
    t6 = b0p**2
    t7 = t3 * t2
    e = (
        e0
        + t1 * t4 / 8.0
        + t5 * v0 * t4 * b0pp / 24.0
        - t1 * t4 * b0p / 8.0
        + t1 * t4 * t6 / 24.0
        + t1 * t7 * b0p / 6.0
        - t1 * t7 / 3.0
        + t1 * t3 / 2.0
    )

    return e


def rPT5(vol, pars):
    """
    Natrual strain EOS with 4 paramters
    Seems only work in near-equillibrium range.
    Implementions as Alberto Otero-de-la-Roza, i.e. rPT5 is used here
    Comput Physics Comm, 2011, 182: 1708-1720,
    in their article, labeled as PT3 (3-order), however, we mention it as
    rPT5 for 4-parameters EOS.
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = pars[4]

    x = vol / v0
    fn = 1 / 3 * np.log(x)
    H = b0 * bpp + bp * bp

    E = e0 + 9 / 8 * b0 * v0 * fn**2 * (
        -3 * (H + 3 * bp - 3) * fn**2 - 4 * (bp - 2) * fn + 4
    )
    return E


def res_rPT5(pars, y, x):
    res = y - rPT5(x, pars)
    return res


def rPT5_pv(vol, pars):
    """
    Natrual strain (Poirier-Tarantola)EOS with 5 paramters
    Implementions as Alberto Otero-de-la-Roza, i.e. rPT5 is used here
    Comput Physics Comm, 2011, 182: 1708-1720,
    in their article, labeled as PT3 (3-order), however, we mention it as
    rPT5 for 4-parameters EOS.
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = pars[4]

    x = vol / v0
    fn = 1 / 3 * np.log(x)
    H = b0 * bpp + bp * bp

    P = (
        -1.5
        * b0
        * fn
        * np.exp(-3 * fn)
        * (-3 * (H + 3 * bp - 3) * fn**2 - 3 * (bp - 2) * fn + 2)
    )
    return P


def res_rPT5_pv(pars, y, x):
    res = y - rPT5_pv(x, pars)
    return res


# ----------------------------------------------------------------------------------------
def res_vinet(pars, y, x):
    res = y - vinet(x, pars)
    return res


def calc_props_vinet(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    bpp = (19 - 18 * bp - 9 * bp * bp) / (36 * b0)
    props = [e0, b0, bp, v0, bpp]
    return props


def vinet(vol, pars):
    """
    Vinet equation from PRB 70, 224107
    Following, Shang Shunli et al., comput mater sci, 2010: 1040-1048, original expressions.
    """
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    eta = (vol / v0) ** (1 / 3)
    Y = 1.5 * (bp - 1) * (1 - eta)
    Z = 4 * b0 * v0 / (bp - 1) ** 2

    ee = e0 + Z - Z * (1 - Y) * np.exp(Y)
    return ee


def vinet_pv(vol, pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    eta = (vol / v0) ** (1 / 3)
    P = 3 * b0 * (1 - eta / eta**2) * np.exp(-1.5 * (bp - 1) * (eta - 1))
    return P


def res_vinet_pv(par, y, x):
    res = y - vinet(x, pars)
    return res


# ----------------------------------------------------------------------------------------
def Li4p(V, parameters):
    """Li JH, APL, 87, 194111 (2005)"""
    E0 = parameters[0]
    B0 = parameters[1]
    BP = parameters[2]
    V0 = parameters[3]

    # In fact, it is just a small modified version of Rose or Vinet EOS.
    x = (V / V0) ** (1.0 / 3.0)
    eta = np.sqrt(-9 * B0 * V0 / E0)
    astar = eta * (x - 1.0)
    delta = (BP - 1.0) / (2 * eta) - 1.0 / 3.0

    E = E0 * (1 + astar + delta * astar**3) * np.exp(-astar)

    return E


def res_Li4p(p, y, x):
    return y - Li4p(x, p)


# ----------------------------------------------------------------------------------------
def morse(v, pars):
    """Reproduce from ShunliShang's matlab script."""
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]

    # usually used in high pressure range, but maybe failed in large
    # expansion( > 2*v0) range.
    a = e0 + (9 * b0 * v0) / (2 * (-1 + bp) * (-1 + bp))
    b = (-9 * b0 * np.exp(-1 + bp) * v0) / (-1 + bp) / (-1 + bp)
    c = (9 * b0 * np.exp(-2 + 2 * bp) * v0) / (2 * (-1 + bp) * (-1 + bp))
    d = (1 - bp) / (v0 ** (1.0 / 3))

    ee = (
        a
        + b * np.exp(d * np.power(v, 1.0 / 3))
        + c * np.exp(2 * d * np.power(v, 1.0 / 3))
    )
    return ee


def calc_props_morse(pars):
    e0 = pars[0]
    b0 = pars[1]
    bp = pars[2]
    v0 = pars[3]
    bpp = (5 - 5 * bp - 2 * bp * bp) / (9 * b0)

    props = [e0, b0, bp, v0, bpp]
    return props


def res_morse(p, en, volume):
    res = en - morse(volume, p)
    return res


def morse_AB(volume, p):
    """
    morse_AB EOS formula from Song's FVT souces
    """
    # p0 = [e0, b0, bp, v0, bpp]
    E0 = p[0]
    A = p[1]
    B = p[2]
    V0 = p[3]
    xx = (volume / V0) ** (1 / 3)
    E = (
        1.5
        * V0
        * A
        * (np.exp(B * (1.0 - 2.0 * xx)) - 2.0 * np.exp(-B * xx) + np.exp(-B))
        + E0
    )
    return E


def res_morse_AB(p, en, volume):
    res = en - morse_AB(volume, p)
    return res


def morse_3p(volume, p):
    """
    morse_AB EOS formula from Song's FVT souces
    A= 0.5*B
    """
    # p0 = [e0, b0, bp, v0, bpp]
    E0 = p[0]
    A = p[1]
    V0 = p[2]
    B = 0.5 * A
    xx = (volume / V0) ** (1 / 3)
    E = (
        1.5
        * V0
        * A
        * (np.exp(B * (1.0 - 2.0 * xx)) - 2.0 * np.exp(-B * xx) + np.exp(-B))
        + E0
    )
    return E


def res_morse_3p(p, en, volume):
    res = en - morse_3p(volume, p)
    return res


def morse_6p(vol, par):
    """
    Generalized Morse EOS proposed by Qin, see:
    Qin et al. Phys Rev B, 2008, 78, 214108.
    Qin et al. Phys Rev B, 2008, 77, 220103(R).
    """
    # p0 = [e0, b0, bp, v0, bpp]
    e0 = par[0]
    b0 = par[1]
    p = abs(par[2])
    v0 = par[3]
    q = abs(par[4])
    m = abs(par[5])
    n = abs(par[6])

    C = (p + m) * (q - n) * ((p + m) - (q - n)) + (p * n + q * m)
    A = 9 * b0 * v0 * (q - n) / C
    B = 9 * b0 * v0 * (p + m) / C

    x = (vol / v0) ** (1 / 3)
    ee = (
        A * np.exp(-p * (x - 1)) / (x**m)
        - B * (x**n) * np.exp(-q * (x - 1))
        + e0
        - (A - B)
    )
    return ee


def calc_props_morse_6p(par):
    e0 = par[0]
    b0 = par[1]
    p = abs(par[2])
    v0 = par[3]
    q = abs(par[4])
    m = abs(par[5])
    n = abs(par[6])

    bp = (p + q) / 3 + 1
    bpp = ((7 - p * q) / 9 - bp) / b0
    props = [e0, b0, bp, v0, bpp, m, n]
    return props


def res_morse_6p(p, en, volume):
    res = en - morse_6p(volume, p)
    return res


# ----------------------------------------------------------------------------------------
def mie(v, p):
    """
    Mie model for song's FVT
    """
    # p0 = [e0, b0, bp, v0, bpp]
    E0 = p[0]
    m = p[1]
    n = p[2]
    V0 = p[3]

    xx = (V0 / v) ** (1 / 3)
    E = E0 / (n - m) * (n * (xx**m) - m * (xx**n))

    return E


def res_mie(p, e, v):
    res = e - mie(v, p)
    return res


def mie_simple(v, p):
    """
    Mie_simple model for song's FVT
    """
    # p0 = [e0, b0, bp, v0, bpp]
    E0 = p[0]
    m = 4
    n = p[2]
    V0 = p[3]

    xx = (V0 / v) ** (1 / 3)
    E = E0 / (n - m) * (n * (xx**m) - m * (xx**n))

    return E


def res_mie_simple(p, e, v):
    res = e - mie_simple(v, p)
    return res


# ----------------------------------------------------------------------------------------
def TEOS(v, par):
    """
    Holland, et al, Journal of Metamorphic Geology, 2011, 29(3): 333-383
    Modified Tait equation of Huang & Chow
    """
    e0 = par[0]
    b0 = par[1]
    bp = par[2]
    v0 = par[3]
    bpp = par[4]

    a = (1 + bp) / (1 + bp + b0 * bpp)
    b = bp / b0 - bpp / (1 + bp)
    c = (1 + bp + b0 * bpp) / (bp**2 + bp - b0 * bpp)

    t1 = (v / v0) + a - 1
    P = 1 / b * ((t1 / a) ** (-1 / c) - 1)
    # ?? the following energy relation is nonsenses.
    E = -P * np.log(v0) + e0
    return E


def res_TEOS(p, e, v):
    res = e - TEOS(v, p)
    return res


# ----------------------------------------------------------------------------------------
def SJX_v2(vol, par):
    """
    Sun Jiuxun, et al. J phys Chem Solids, 2005, 66: 773-782.
    They said it is satified for the limiting condition at high pressure.
    """
    e0 = par[0]
    b0 = par[1]
    bp = par[2]
    v0 = par[3]

    Y = (v0 / vol) ** (1 / 3) - 1
    alpha = 1 / 4 * (3 * bp - 1)
    beta = 1 - 1 / alpha
    a = alpha
    b = beta
    ee = e0 + 3 / 10 * a**4 * b0 * v0 * (
        2 * (b**5 + 1 / a**5) * (1 - (Y + 1) ** (-3))
        - 15 * b**4 * (1 - (Y + 1) ** (-2))
        - 60 * b**3 * (1 - (Y + 1) ** (-1))
        + 60 * b**2 * np.log(Y + 1)
        + 3 * Y * (Y + 2)
        - 30 * b * Y
    )
    return ee


def res_SJX_v2(p, e, v):
    res = e - SJX_v2(v, p)
    return res


def SJX_5p(vol, par):
    """
    SJX_5p's five parameters EOS, Physica B: Condens Mater, 2011, 406: 1276-1282
    """
    e0 = par[0]
    a = par[1]
    b = par[2]
    v0 = par[3]
    n = par[4]

    X = (vol / v0) ** (1 / 3)
    C1 = n * b * np.exp(a * (1 - X) + b * (1 - X**n))
    C2 = (-n * b - a) * np.exp(b * (1 - X**n))
    ee = e0 / a * (C1 + C2)
    return ee


def res_SJX_5p(p, e, v):
    res = e - SJX_5p(v, p)
    return res


def calc_props_SJX_5p(par):
    e0 = par[0]
    a = par[1]
    b = par[2]
    v0 = par[3]
    n = par[4]

    b0 = n * b * e0 / (9 * v0) * (a + n * b + n - 1)
    Q = (a + n * b) * (a + 2 * n * b) - (n - 1) * (n - 2)
    bp = 1 + (Q / 3) / (a + n * b + n - 1)

    props = [e0, b0, bp, v0, n]
    return props


# ----------------------------------------------------------------------------------------
# some utility functions


def read_ve(fin):
    if not os.path.exists(fin):
        print("Could not find input file: [%s]" % fin)
        os.sys.exit(-1)
    lines = open(fin).readlines()
    nline = len(lines)
    vol = []
    en = []
    for i in range(nline):
        tmp = lines[i].split()
        if len(tmp) >= 2:
            # now, we can read ve.dat & vec.dat
            # read the first two numbers.
            v = float(tmp[0])
            e = float(tmp[1])
            vol.append(v)
            en.append(e)
        else:
            pass
    return [vol, en]


def read_vlp(fin, fstart, fend):
    if not os.path.exists(fin):
        print(">> Could not find input file: [%s]" % fin)
        os.sys.exit(-1)
    lines = open(fin).readlines()
    nline = len(lines)
    print("\n** Totally, there are %d data points reading..." % nline)
    vol = []
    cella = []
    cellb = []
    cellc = []
    cellba = []
    cellca = []
    for i in range(nline):
        tmp = lines[i].split()
        if len(tmp) >= 2:
            # read the DATA information
            v = float(tmp[0])
            if len(tmp) == 6:
                # for vlp.dat
                a = float(tmp[1])
                b = float(tmp[2])
                c = float(tmp[3])
                ba = float(tmp[4])
                ca = float(tmp[5])
            elif len(tmp) == 7:
                # for velp.dat
                e = float(tmp[1])
                a = float(tmp[2])
                b = float(tmp[3])
                c = float(tmp[4])
                ba = float(tmp[5])
                ca = float(tmp[6])
            vol.append(v)
            cella.append(a)
            cellb.append(b)
            cellc.append(c)
            cellba.append(ba)
            cellca.append(ca)
    print("\n** Vmin = %f, Vmax = %f" % (min(vol), max(vol)))

    # some special conditions
    if fstart <= 0:
        print("\n** Data range input parameters must be positive values!")
        os.sys.exit(-1)
    if fstart > fend:
        if fend != -1:
            tmp = fstart
            fstart = fend
            fend = tmp
    if fend > nline:
        print(
            "\n** EoSfit fit range exceed available data numbers, Reset it to be %d now."
            % nline
        )
        fend = nline
    if fstart > nline:
        print(
            "EoSfit fit range exceed available data numbers, Reset it to be 1: %d now."
            % nline
        )
        fstart = 1
        fend = -1

    fstart = fstart - 1

    # get data in the predefined range.
    fstart = fstart - 1
    if fend != -1:
        vol = vol[fstart:fend]
        cella = cella[fstart:fend]
        cellb = cellb[fstart:fend]
        cellc = cellc[fstart:fend]
        cellba = cellba[fstart:fend]
        cellca = cellca[fstart:fend]
    else:
        vol = vol[fstart:]
        cella = cella[fstart:]
        cellb = cellb[fstart:]
        cellc = cellc[fstart:]
        cellba = cellba[fstart:]
        cellca = cellca[fstart:]
    return [vol, cella, cellb, cellc, cellba, cellca]


def read_velp(fin, fstart, fend):
    if not os.path.exists(fin):
        print(">> Could not find input file: [%s]" % fin)
        os.sys.exit(-1)
    lines = open(fin).readlines()
    nline = len(lines)
    print("\n** Totally, there are %d data points reading..." % nline)
    vol = []
    eng = []
    cella = []
    cellb = []
    cellc = []
    cellba = []
    cellca = []
    for i in range(nline):
        tmp = lines[i].split()
        if len(tmp) >= 2:
            # read the DATA information
            v = float(tmp[0])
            if len(tmp) == 7:
                # for velp.dat
                e = float(tmp[1])
                a = float(tmp[2])
                b = float(tmp[3])
                c = float(tmp[4])
                ba = float(tmp[5])
                ca = float(tmp[6])
            vol.append(v)
            eng.append(e)
            cella.append(a)
            cellb.append(b)
            cellc.append(c)
            cellba.append(ba)
            cellca.append(ca)
    print("\n** Vmin = %f, Vmax = %f" % (min(vol), max(vol)))

    # some special conditions
    if fstart <= 0:
        print("\n** Data range input parameters must be positive values!")
        os.sys.exit(-1)
    if fstart > fend:
        if fend != -1:
            tmp = fstart
            fstart = fend
            fend = tmp
    if fend > nline:
        print(
            "\n** EoSfit fit range exceed available data numbers, Reset it to be %d now."
            % nline
        )
        fend = -1
    if fstart > nline:
        print(
            "EoSfit fit range exceed available data numbers, Reset it to be 1: %d now."
            % nline
        )
        fstart = 1
        fend = -1

    # get data in the predefined range.
    fstart = fstart - 1
    if fend != -1:
        vol = vol[fstart:fend]
        eng = eng[fstart:fend]
        cella = cella[fstart:fend]
        cellb = cellb[fstart:fend]
        cellc = cellc[fstart:fend]
        cellba = cellba[fstart:fend]
        cellca = cellca[fstart:fend]
    else:
        vol = vol[fstart:]
        eng = eng[fstart:]
        cella = cella[fstart:]
        cellb = cellb[fstart:]
        cellc = cellc[fstart:]
        cellba = cellba[fstart:]
        cellca = cellca[fstart:]

    return [vol, eng, cella, cellb, cellc, cellba, cellca]


def init_guess(fin):
    v, e = read_ve(fin)
    a, b, c = np.polyfit(v, e, 2)  # this comes from pylab
    # initial guesses.
    v0 = np.abs(-b / (2 * a))
    e0 = a * v0**2 + b * v0 + c
    b0 = 2 * a * v0
    bp = 3.0
    bpp = 1 * eV2GPa
    x0 = [e0, b0, bp, v0, bpp]
    return x0


def repro_ve(func, vol_i, p):
    ndata = len(vol_i)
    eni = np.zeros(ndata)
    for i in range(ndata):
        eni[i] = eval(func)(vol_i[i], p)
    return eni


def repro_vp(func, vol_i, pars):
    dv = 1e-5
    efunc = eval(func)
    ndata = len(vol_i)
    Press = np.zeros(ndata)
    for i in range(ndata):
        v = vol_i[i]
        P = (efunc(v + 0.5 * dv, pars) - efunc(v - 0.5 * dv, pars)) / dv
        Press[i] = -P * eV2GPa
    return Press


def ext_vec(
    func, fin, p0, fs, fe, vols=None, vole=None, ndata=101, refit=0, show_fig=False
):
    """
    extrapolate the data points for E-V based on the fitted parameters in small or
    very large volume range.
    """
    # read fitted-parameters
    # pars = np.loadtxt(fin, dtype=float)
    fout = "EoSfit.out"
    pars = lsqfit_eos(func, fin, p0, fs, fe, fout, refit=refit)

    vol, eng, cella, cellb, cellc, cellba, cellca = read_velp(fin, fs, fe)
    sca = ext_splint(vol, cellca)

    # define extrapolate range
    print("\n** Vext_start = %f, Vext_end = %f, N_ext = %d" % (vols, vole, ndata))
    vol_ext = np.linspace(vols, vole, ndata)
    en_ext = eval(func)(vol_ext, pars)

    # en_ext = np.zeros(ndata)
    fout = "ext_ve_" + func + ".dat"
    fw = open(fout, "w+")
    fw.write("%d\n" % ndata)
    for i in range(ndata):
        vx = vol_ext[i]
        ex = en_ext[i]
        if vx <= min(vol):
            cax = cellca[0]
        elif vx >= max(vol):
            cax = cellca[-1]
        else:
            cax = sca(vx)
        fw.write("%f\t%f\t%f\n" % (vx, ex, cax))
        fw.flush()
    fw.close()

    # plot the results
    vol, en = read_ve(fin)
    plt.plot(vol, en, "o-", vol_ext, en_ext, "rd")
    plt.legend(["dft_calc", func + "_ext"], loc="best")
    plt.savefig("ext_ve_" + func + ".png")
    if show_fig:
        plt.show()
    plt.close()
    print("\n>> Storing the extrapolate results in %s\n" % fout)
    print("\n>> DONE!")

    return


def ext_splint(xp, yp, order=3, method="unispl"):
    if method == "interp1d":
        # old-implement of 1-D interpolation
        # could not used in extrapolate
        SPLINT = interp1d
        return SPLINT(xp, yp, order, bounds_error=False)
    elif method == "piecepoly":
        SPLINT = PiecewisePolynomial
        return SPLINT(xp, yp, order)
    else:
        if method == "unispl":
            # 1-D smoothing spline fit to a given set of data
            SPLINT = UnivariateSpline
        else:
            SPLINT = LSQUnivariateSpline
        return SPLINT(xp, yp, k=order)


def ext_velp(
    fin,
    fstart,
    fend,
    vols,
    vole,
    ndata,
    order=3,
    method="unispl",
    fout="ext_velp.dat",
    show_fig=False,
):
    """
    extrapolate the lattice parameters based on input data
    """
    # read file
    vol, eng, cella, cellb, cellc, cellba, cellca = read_velp(fin, fstart, fend)

    # define extrapolate range
    print("\n** Vext_start = %f, Vext_end = %f, N_ext = %d" % (vols, vole, ndata))
    vv = np.linspace(vols, vole, ndata)

    # spline order = 3 by default
    se = ext_splint(vol, eng, order, method)
    ee = se(vv)
    sa = ext_splint(vol, cella, order, method)
    cellaa = sa(vv)
    sb = ext_splint(vol, cellb, order, method)
    cellbb = sb(vv)
    sc = ext_splint(vol, cellc, order, method)
    cellcc = sc(vv)
    sba = ext_splint(vol, cellba, order, method)
    cellbaba = sba(vv)
    sca = ext_splint(vol, cellca, order, method)
    cellcaca = sca(vv)
    cellca_cal = cellcc / cellaa

    # plot the extrapolate results
    nfigure = 6
    lp_ext = [ee, cellaa, cellbb, cellcc, cellbaba, cellcaca, cellca_cal]
    lp_ori = [
        eng,
        cella,
        cellb,
        cellc,
        cellba,
        cellca,
        np.array(cellc) / np.array(cella),
    ]
    lp_ylabel = ["E(eV)", "a(A)", "b(A)", "c(A)", "b/a", "c/a", "c_ext/a_ext"]
    for i in range(nfigure):
        plt.subplot(nfigure, 1, i + 1)
        plt.ylabel(lp_ylabel[i])
        plt.plot(vol, lp_ori[i], "o-", vv, lp_ext[i], "rd")
        if i == 0:
            plt.legend(["ori", "ext"], loc="best")
    plt.xlabel("Volume (A**3)")
    plt.savefig("ext_velp.png")
    if show_fig:
        plt.show()
    plt.close()

    # save the fit data
    # get vba.dat
    fw = open(fout, "w+")
    fw.write(
        "#%12s\t%12s\t%12s\t%12s\t%12s\t%12s\t%12s\t%12s\n"
        % ("volume", "eng", "cella", "cellb", "cellc", "b/a", "c/a", "cext/aext")
    )
    for i in range(ndata):
        fw.write(
            "%12.6f\t%12.6f\t%12.6f\t%12.6f\t%12.6f\t%12.6f\t%12.6f\t%12.6f\n"
            % (
                vv[i],
                ee[i],
                cellaa[i],
                cellbb[i],
                cellcc[i],
                cellbaba[i],
                cellcaca[i],
                cellca_cal[i],
            )
        )
        fw.flush()
    fw.close()
    print("\n>> Storing the extrapolate results in %s\n" % fout)
    print("\n>> DONE!")
    return


def lsqfit_eos(
    func, fin, par, fstart, fend, show_fig=False, fout="EoSfit.out", refit=-1
):
    # make the screen output better.
    print("\n")
    print("\t>> We are using [ %s ] to fit the V-E relationship << \t" % func)
    print("\n")

    fs = fstart
    fe = fend
    p = par

    if fs < 0:
        print("start fitting range index must be a positive integer!")
        os.sys.exit(-1)

    # p0 = [e0, b0, bp, v0, bpp]
    if refit == -1:
        if (func == "morse_AB") or (func == "morse_3p"):
            A = 6
            B = 0.5 * A
            if func == "morse_AB":
                p0 = [p[0], A, B, p[3]]
            else:
                p0 = [p[0], A, p[3]]
        elif func in ["mie", "mie_simple"]:
            p0 = [p[0], 4, 6, p[3]]
        elif func == "morse_6p":
            P = 1
            Q = 2
            m = 1
            n = 1
            p0 = [p[0], p[1], P, p[3], Q, m, n]
        elif func == "SJX_5p":
            alpha = 1
            beta = 1
            n = 1
            p0 = [p[0], alpha, beta, p[3], n]
        else:
            p0 = p
        print(
            ">> use initial guess of fitted-paramters by polynomial fitting results:\n"
        )
    else:
        p0 = np.loadtxt(fout)
        print(">> use initial guess of fitted-paramters by last fitting results:\n")
    print(p0)

    vol, en = read_ve(fin)
    ndata = len(vol)
    # To determine the data points included in the fitting
    if fe > ndata:
        fe = ndata
        print(
            "\n[WARNING]: f_end exceed available data numbers, Reset it to be %d now."
            % ndata
        )
    if fs > ndata:
        print(
            "\n[WARNING]: f_start exceed available data numbers, Reset it to be 1: %d now."
            % ndata
        )
        print("and Reset f_end to be %d now." % ndata)
        fs = 1
        fe = ndata

    if fe == -1:
        print("\n[ATTENTIONS]: fend = -1")
        num = ndata - fs + 1
    else:
        num = fe - fs + 1

    vol = vol[fs - 1 : fe]
    en = en[fs - 1 : fe]
    print("\n%d/%d data was used in the fitting ...\n" % (num, ndata))

    # *************************************************************************
    # fit it: step 1.
    popt, pcov, infodict, mesg, ier = leastsq(
        eval("res_" + func),
        p0,
        args=(en, vol),
        full_output=1,
        maxfev=(len(p0) + 1) * 400,
    )
    nfev = infodict["nfev"]
    fvec = infodict["fvec"]

    """
    psi_i = sum(fvec**2)
    psi_min = min(fvec)
    pfactor = np.zeros(num)
    wi = (4 + 1) / num * psi_i / psi_min
    pfi = np.exp(-wi * wi)
    """
    if ier not in [1, 2, 3, 4]:
        raise RuntimeError("Optimal parameters not found: " + mesg)
    # *************************************************************************

    print("*" * 80)
    print(">> fitted parameters (with %d iterations):" % nfev)
    # p0 = [e0, b0, bp, v0, bpp]
    if func == "morse_AB":
        e0, A, B, v0 = popt
        print("%12s\t%12s\t%12s\t%12s" % ("V0(A**3)", "A", "B", "E0(eV)"))
        print("%12f\t%12f\t%12f\t%12f\n" % (v0, A, B, e0))
    elif func == "morse_3p":
        e0, A, v0 = popt
        B = 0.5 * A
        print("%12s\t%12s\t%12s\t%12s" % ("V0(A**3)", "A", "B", "E0(eV)"))
        print("%12f\t%12f\t%12f\t%12f\n" % (v0, A, B, e0))
    elif func in ["mie", "mie_simple"]:
        e0, m, n, v0 = popt
        print("%12s\t%12s\t%12s\t%12s" % ("V0(A**3)", "m", "n", "E0(eV)"))
        print("%12f\t%12f\t%12f\t%12f\n" % (v0, m, n, e0))
    elif func == "morse_6p":
        e0, b0, bp, v0, bpp, m, n = calc_props_morse_6p(popt)
        b0 = eV2GPa * b0
        bpp = bpp / eV2GPa
        print(
            "%12s\t%12s\t%12s\t%12s\t%12s\t%12s\t%12s"
            % ("V0(A**3)", "B0(GPa)", "Bp", "E0(eV)", "Bpp(1/GPa)", "m", "n")
        )
        print(
            "%12f\t%12f\t%12f\t%12f\t%12f\t%12f\t%12f\n" % (v0, b0, bp, e0, bpp, m, n)
        )
    elif func == "SJX_5p":
        e0, b0, bp, v0, n = calc_props_SJX_5p(popt)
        b0 = eV2GPa * b0
        print(
            "%12s\t%12s\t%12s\t%12s\t%12s"
            % ("V0(A**3)", "B0(GPa)", "Bp", "E0(eV)", "n")
        )
        print("%12f\t%12f\t%12f\t%12f\t%12f\n" % (v0, b0, bp, e0, n))
    elif func in ["mBM4poly", "mBM5poly", "mBM4", "LOG4", "vinet", "morse", "BM4"]:
        prop_func = eval("calc_props_" + func)
        e0, b0, bp, v0, bpp = prop_func(popt)
        b0 = eV2GPa * b0
        bpp = bpp / eV2GPa
        print(
            "%12s\t%12s\t%12s\t%12s\t%12s"
            % ("V0(A**3)", "B0(GPa)", "Bp", "E0(eV)", "Bpp(1/GPa)")
        )
        print("%12f\t%12f\t%12f\t%12f\t%12f\n" % (v0, b0, bp, e0, bpp))
    else:
        e0, b0, bp, v0, bpp = popt
        b0 = eV2GPa * b0
        bpp = bpp / eV2GPa
        print(
            "%12s\t%12s\t%12s\t%12s\t%12s"
            % ("V0(A**3)", "B0(GPa)", "Bp", "E0(eV)", "Bpp(1/GPa)")
        )
        print("%12f\t%12f\t%12f\t%12f\t%12f\n" % (v0, b0, bp, e0, bpp))

    # write the fitted results in fit.out
    fw = open(fout, "w+")
    for i in range(len(popt)):
        fw.write("%f\n" % popt[i])
        fw.flush()
    fw.close()

    # data evaluate by fitted parameters --to be used in the plot
    vol_i = np.linspace(min(vol) * 0.95, max(vol) * 1.05, len(vol) * 2)
    en_i = repro_ve(func, vol_i, popt)

    print("*" * 80)
    # calculate fitted residuals and square errors
    res_opt = np.zeros(len(vol))
    for i in range(len(vol)):
        res_opt[i] = (fvec[i]) ** 2
    fit_res = sum(res_opt)
    fit_var = np.var(fvec)
    fit_std = np.std(fvec)
    print("\nfitted residuals\t= %16e\n" % fit_res)
    print("fitted variations\t= %16e\n" % fit_var)
    print("standard deviations\t= %16e\n" % fit_std)
    # if fit_res > 1e-4:
    #    print("\n>> Residuals seems too large, please refit it by swithing argument --refit 1!\n")
    #    show = 'F'  # reset show tag, not to show the figure.
    plt.plot(vol, en, "o", vol_i, en_i)
    plt.title("EoS fitted by: %s model" % str(func))
    plt.legend(["calc", func + "-fit"], loc="best")
    plt.xlabel("Volume (A**3)")
    plt.ylabel("Energy (eV)")
    plt.savefig("EoSfit_" + func + ".png")
    if show_fig:
        plt.show()
    print("*" * 80)
    plt.close()

    # reproduce data by the popt and write it into files.
    repro_en = repro_ve(func, vol, popt)
    repro_press = repro_vp(func, vol, popt)
    fve = open(func + "_ve_fit.dat", "w+")
    fvp = open(func + "_vp_fit.dat", "w+")
    fve.write(
        "#%20s\t%20s\t%20s\t%20s\n"
        % ("volume(A**3)", "energy(fit)", "energy(cal)", "dE(%)")
    )
    fvp.write(
        "#%20s\t%20s\t%20s\t%20s\n"
        % ("volume(A**3)", "pressure(GPa)", "pressure(Mbar)", "pressure(kbar)")
    )
    for i in range(len(vol)):
        fve.write(
            "%20f\t%20f\t%20f\t%20f\n"
            % (vol[i], repro_en[i], en[i], 100 * np.abs((en[i] - repro_en[i]) / en[i]))
        )
        fve.flush()
        p_tmp = repro_press[i]
        fvp.write("%20f\t%20f\t%20f\t%20f\n" % (vol[i], p_tmp, p_tmp / 100, p_tmp * 10))
        fvp.flush()
    fve.close()
    fvp.close()
    return popt


def parse_argument():
    parser = argparse.ArgumentParser(
        description=" Script to fit EOS in MTP calculations"
    )
    parser.add_argument("choice", help="to run mfp, ext_vec, ext_velp?")
    parser.add_argument("infile", help="input ve.dat, vec.dat, velp.dat, or vlp.dat")
    parser.add_argument(
        "-eos",
        "--eos",
        default="morse_AB",
        help="Please choose one of the EOS name in: " + str(get_eos_list()),
    )
    parser.add_argument(
        "--show", default="T", help="to show the fitted results[T] or not[F]?"
    )
    parser.add_argument(
        "-vr",
        "--vrange",
        type=float,
        nargs=3,
        default=[30, 50, 101],
        help="extend range used for ext, e.g. [vols, vole, ndata]: 1 10 31",
    )
    parser.add_argument(
        "-fr",
        "--frange",
        type=int,
        nargs=2,
        default=[1, 1000],
        help="data range to be fitted, e.g. [ns, ne]",
    )
    parser.add_argument(
        "--refit",
        type=int,
        default=-1,
        help="fit the data with last fitted parameters as initial guess",
    )
    parser.add_argument(
        "-eord",
        "--extorder",
        type=int,
        default=3,
        help="spline order used in extrapolate",
    )
    parser.add_argument(
        "-enum",
        "--extnum",
        type=int,
        default=20,
        help="number of data to be extrapolate",
    )
    parser.add_argument(
        "-v", "--version", action="version", version="%(prog)s " + __version__()
    )
    args = parser.parse_args()
    return args


# main
if __name__ == "__main__":
    args = parse_argument()
    cho = args.choice
    fin = args.infile
    func = args.eos
    SHOW = args.show
    REFIT = args.refit

    eos_list = get_eos_list()
    if func not in eos_list:
        print("Choose one of the following eos name:")
        print(eos_list)
        os.sys.exit(-1)
    else:
        pass
    # p0 = [e0, b0, bp, v0, bpp]

    # used in mfp choice
    fs = args.frange[0]
    fe = args.frange[1]

    if fs < 0:
        print("ERROR, start range index must be a positive integer!")
        os.sys.exit(-1)

    if cho == "mfp":
        p0 = init_guess(fin)
        lsqfit_eos(func, fin, p0, fs, fe, SHOW, refit=REFIT)
    elif cho == "ext_vec":
        # used in ext choice
        ExtVS = args.vrange[0]
        ExtVE = args.vrange[1]
        ExtNum = int(args.vrange[2])
        if ExtVS < 0 or ExtVE < 0 or ExtNum < 0:
            print("ERROR, range setting must be a positive value!")
            exit(1)
        # if ExtNum % 2 == 0:
        #    print("[WARNING]: ndata = %d, reset it to be %d" %
        #          (ExtNum, ExtNum + 1))
        #    ExtNum = ExtNum + 1
        p0 = init_guess(fin)
        ext_vec(func, fin, p0, fs, fe, ExtVS, ExtVE, ExtNum, refit=REFIT)
    elif cho == "ext_velp":
        EORDER = args.extorder
        ExtVS = args.vrange[0]
        ExtVE = args.vrange[1]
        ExtNum = int(args.vrange[2])
        if ExtVS < 0 or ExtVE < 0 or ExtNum < 0:
            print("ERROR, range setting must be a positive value!")
            exit(1)
        if ExtNum % 2 == 0:
            print("[WARNING]: ndata = %d, reset it to be %d" % (ExtNum, ExtNum + 1))
            ExtNum = ExtNum + 1
        ext_velp(fin, fs, fe, ExtVS, ExtVE, ExtNum, order=EORDER, method="unispl")
    else:
        print("Unkown Choice, abort now ... ...")
        os.sys.exit(-1)



import glob
import os
import shutil
import sys
import unittest

import numpy as np
from monty.serialization import loadfn

from apex.core.calculator.ABACUS import ABACUS
from apex.core.calculator.lib import abacus_scf

from apex.core.common_prop import make_property
from apex.core.property.Elastic import Elastic
from apex.core.property.EOS import EOS
from apex.core.property.Gamma import Gamma
from apex.core.property.Interstitial import Interstitial
from apex.core.property.Surface import Surface
from apex.core.property.Vacancy import Vacancy

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"


class TestABACUS(unittest.TestCase):
    def setUp(self):
        self.jdata = {
            "structures": ["confs/fcc-Al"],
            "interaction": {
                "type": "abacus",
                "incar": "abacus_input/INPUT",
                "potcar_prefix": "abacus_input",
                "potcars": {"Al": "Al_ONCV_PBE-1.0.upf"},
                "orb_files": {"Al": "Al_gga_9au_100Ry_4s4p1d.orb"},
            },
        }

        self.conf_path = "confs/fcc-Al"
        self.equi_path = "confs/fcc-Al/relaxation/relax_task"
        self.source_path = "equi/abacus"
        if not os.path.exists(self.equi_path):
            os.makedirs(self.equi_path)
        if not os.path.exists(os.path.join(self.equi_path, "OUT.ABACUS")):
            os.makedirs(os.path.join(self.equi_path, "OUT.ABACUS"))
        for ifile in ["INPUT", "STRU"]:
            if not os.path.exists(os.path.join(self.equi_path, ifile)):
                shutil.copy(
                    os.path.join(self.source_path, ifile),
                    os.path.join(self.equi_path, ifile),
                )
        for ifile in ["running_cell-relax.log", "STRU_ION_D"]:
            if not os.path.exists(os.path.join(self.equi_path, "OUT.ABACUS", ifile)):
                shutil.copy(
                    os.path.join(self.source_path, ifile),
                    os.path.join(self.equi_path, "OUT.ABACUS", ifile),
                )
        shutil.copy(
            os.path.join(self.source_path, "cell-relax.json"),
            os.path.join(self.equi_path, "result.json"),
        )

        self.confs = self.jdata["structures"]
        self.inter_param = self.jdata["interaction"]
        self.ABACUS = ABACUS(self.inter_param, os.path.join(self.conf_path, "STRU"))

    def tearDown(self):
        if os.path.exists("confs/fcc-Al/relaxation"):
            shutil.rmtree("confs/fcc-Al/relaxation")
        if os.path.exists("confs/fcc-Al/eos_00"):
            shutil.rmtree("confs/fcc-Al/eos_00")
        if os.path.exists("confs/fcc-Al/eos_02"):
            shutil.rmtree("confs/fcc-Al/eos_02")
        if os.path.exists("confs/fcc-Al/elastic_00"):
            shutil.rmtree("confs/fcc-Al/elastic_00")
        if os.path.exists("confs/fcc-Al/vacancy_00"):
            shutil.rmtree("confs/fcc-Al/vacancy_00")
        if os.path.exists("confs/fcc-Al/interstitial_00"):
            shutil.rmtree("confs/fcc-Al/interstitial_00")
        if os.path.exists("confs/fcc-Al/surface_00"):
            shutil.rmtree("confs/fcc-Al/surface_00")
        if os.path.exists("confs/fcc-Al/gamma_00"):
            shutil.rmtree("confs/fcc-Al/gamma_00")

    def test_make_property(self):
        property = {"type": "eos", "vol_start": 0.85, "vol_end": 1.15, "vol_step": 0.01}
        make_property(self.jdata["structures"], self.jdata["interaction"], [property])
        self.assertTrue(os.path.exists(os.path.join(self.conf_path, "eos_00")))
        self.assertTrue(os.path.exists(os.path.join(self.conf_path, "eos_00", "INPUT")))
        for ii in glob.glob(os.path.join(self.conf_path, "eos_00", "task.*")):
            self.assertTrue(os.path.exists(os.path.join(ii, "INPUT")))
            self.assertTrue(os.path.exists(os.path.join(ii, "pp_orb")))
            self.assertTrue(os.path.exists(os.path.join(ii, "KPT")))
            self.assertTrue(os.path.exists(os.path.join(ii, "STRU")))
            self.assertEqual(
                os.path.realpath(os.path.join(ii, "pp_orb", "Al_ONCV_PBE-1.0.upf")),
                os.path.realpath(
                    os.path.join(
                        self.jdata["interaction"]["potcar_prefix"],
                        "Al_ONCV_PBE-1.0.upf",
                    )
                ),
            )

    def test_make_property_eos(self):
        property = {"type": "eos", "vol_start": 0.85, "vol_end": 1.15, "vol_step": 0.01}
        work_path = os.path.join(self.conf_path, "eos_00")
        eos = EOS(property, self.inter_param)
        eos.make_confs(work_path, self.equi_path, refine=False)

        for ii in glob.glob(os.path.join(work_path, "task.*")):
            self.assertTrue(os.path.isfile(os.path.join(ii, "STRU")))
            self.assertTrue(os.path.isfile(os.path.join(ii, "eos.json")))
            self.assertEqual(
                os.path.realpath(os.path.join(ii, "STRU.orig")),
                os.path.realpath(
                    os.path.join(self.equi_path, "OUT.ABACUS", "STRU_ION_D")
                ),
            )

            eos_json = loadfn(os.path.join(ii, "eos.json"))
            stru_data = abacus_scf.get_abacus_STRU(
                os.path.realpath(os.path.join(ii, "STRU"))
            )
            vol_per_atom = (
                abs(np.linalg.det(stru_data["cells"]))
                / np.array(stru_data["atom_numbs"]).sum()
            )
            self.assertAlmostEqual(eos_json["volume"], vol_per_atom)

    def test_make_property_elastic(self):
        property = {"type": "elastic", "norm_deform": 1e-2, "shear_deform": 1e-2}
        work_path = os.path.join(self.conf_path, "elastic_00")
        elastic = Elastic(property, self.inter_param)
        elastic.make_confs(work_path, self.equi_path, refine=False)

        self.assertEqual(
            os.path.realpath(os.path.join(work_path, "STRU")),
            os.path.realpath(os.path.join(self.equi_path, "OUT.ABACUS", "STRU_ION_D")),
        )
        for ii in glob.glob(os.path.join(work_path, "task.*")):
            self.assertTrue(os.path.isfile(os.path.join(ii, "STRU")))
            self.assertTrue(os.path.isfile(os.path.join(ii, "strain.json")))

        os.remove(
            os.path.realpath(os.path.join(self.equi_path, "OUT.ABACUS", "STRU_ION_D"))
        )
        with self.assertRaises(RuntimeError):
            elastic.make_confs(work_path, self.equi_path, refine=False)

    def test_make_property_elastic_post_process(self):
        property = {"type": "elastic", "norm_deform": 1e-2, "shear_deform": 1e-2}
        make_property(self.jdata["structures"], self.jdata["interaction"], [property])
        work_path = os.path.join(self.conf_path, "elastic_00")

        self.assertTrue(os.path.exists(os.path.join(work_path, "INPUT")))
        self.assertTrue(os.path.exists(os.path.join(work_path, "KPT")))

        for ii in glob.glob(os.path.join(work_path, "task.*")):
            self.assertEqual(
                os.path.realpath(os.path.join(work_path, "KPT")),
                os.path.realpath(os.path.join(ii, "KPT")),
            )
            self.assertEqual(
                os.path.realpath(os.path.join(work_path, "INPUT")),
                os.path.realpath(os.path.join(ii, "INPUT")),
            )

    def test_make_property_vacancy(self):
        property = {"type": "vacancy", "supercell": [1, 1, 1]}
        work_path = os.path.join(self.conf_path, "vacancy_00")
        vacancy = Vacancy(property, self.inter_param)
        vacancy.make_confs(work_path, self.equi_path, refine=False)

        self.assertEqual(
            os.path.realpath(os.path.join(work_path, "STRU")),
            os.path.realpath(os.path.join(self.equi_path, "OUT.ABACUS", "STRU_ION_D")),
        )

        stru_data = abacus_scf.get_abacus_STRU(
            os.path.realpath(os.path.join(work_path, "STRU"))
        )
        natom1 = np.array(stru_data["atom_numbs"]).sum()
        for ii in glob.glob(os.path.join(work_path, "task.*")):
            self.assertTrue(os.path.isfile(os.path.join(ii, "STRU")))
            stru_data = abacus_scf.get_abacus_STRU(
                os.path.realpath(os.path.join(ii, "STRU"))
            )
            natom2 = np.array(stru_data["atom_numbs"]).sum()
            self.assertTrue(natom1 == natom2 + 1)

    def test_make_property_interstitial(self):
        property = {"type": "interstitial", "supercell": [2, 2, 2], "insert_ele": ["H"]}
        self.inter_param["potcars"]["H"] = "H_ONCV_PBE-1.0.upf"
        self.inter_param["orb_files"]["H"] = "H_gga_8au_100Ry_2s1p.orb"

        work_path = os.path.join(self.conf_path, "interstitial_00")
        if os.path.exists(work_path):
            shutil.rmtree(work_path)
        os.makedirs(work_path)
        interstitial = Interstitial(property, self.inter_param)
        interstitial.make_confs(work_path, self.equi_path, refine=False)

        self.assertEqual(
            os.path.realpath(os.path.join(work_path, "STRU")),
            os.path.realpath(os.path.join(self.equi_path, "OUT.ABACUS", "STRU_ION_D")),
        )
        stru_data = abacus_scf.get_abacus_STRU(
            os.path.realpath(os.path.join(work_path, "STRU"))
        )
        supercell_scale = property["supercell"][0] * \
                          property["supercell"][1] * \
                          property["supercell"][2]
        natom1 = np.array(stru_data["atom_numbs"]).sum() * supercell_scale
        for ii in glob.glob(os.path.join(work_path, "task.*")):
            self.assertTrue(os.path.isfile(os.path.join(ii, "STRU")))
            stru_data = abacus_scf.get_abacus_STRU(
                os.path.realpath(os.path.join(ii, "STRU"))
            )
            self.assertTrue("H" in stru_data["atom_names"])
            natom2 = np.array(stru_data["atom_numbs"]).sum()
            self.assertTrue(natom1 == natom2 - 1)

    def test_make_property_surface(self):
        property = {
            "type": "surface",
            "min_slab_size": 15,
            "min_vacuum_size": 11,
            "pert_xz": 0.01,
            "max_miller": 2,
            "cal_type": "static",
        }
        work_path = os.path.join(self.conf_path, "surface_00")
        surface = Surface(property, self.inter_param)
        surface.make_confs(work_path, self.equi_path, refine=False)

        self.assertEqual(
            os.path.realpath(os.path.join(work_path, "STRU")),
            os.path.realpath(os.path.join(self.equi_path, "OUT.ABACUS", "STRU_ION_D")),
        )
        for ii in glob.glob(os.path.join(work_path, "task.*")):
            self.assertTrue(os.path.isfile(os.path.join(ii, "STRU")))
            self.assertTrue(os.path.isfile(os.path.join(ii, "miller.json")))

    def test_make_property_gamma(self):
        property = {
            "type":            "gamma",
            "plane_miller":    [0,0,1],
            "slip_direction":  [1,0,0],
            "hcp": {
                    "plane_miller":    [0,0,0,1],
                    "slip_direction":  [2,-1,-1,0],
                    },
            "supercell_size":   [1,1,10],
            "vacuum_size": 10,
            "add_fix": ["true","true","false"],
            "n_steps":         10
        }
        work_path = os.path.join(self.conf_path, "gamma_00")
        gamma = Gamma(property, self.inter_param)
        gamma.make_confs(work_path, self.equi_path, refine=False)

        dfm_dirs = glob.glob(os.path.join(work_path, "task.*"))
        self.assertEqual(len(dfm_dirs), gamma.n_steps + 1)

        self.assertEqual(
            os.path.realpath(os.path.join(work_path, "STRU")),
            os.path.realpath(os.path.join(self.equi_path, "OUT.ABACUS", "STRU_ION_D")),
        )
        for ii in glob.glob(os.path.join(work_path, "task.*")):
            self.assertTrue(os.path.isfile(os.path.join(ii, "STRU")))
            self.assertTrue(os.path.isfile(os.path.join(ii, "miller.json")))

    def test_make_property_refine(self):
        property = {"type": "eos", "vol_start": 0.85, "vol_end": 1.15, "vol_step": 0.01}
        pwd = os.getcwd()
        target_path_0 = "confs/fcc-Al/eos_00"
        target_path_2 = "confs/fcc-Al/eos_02"
        path_to_work = os.path.abspath(target_path_0)

        make_property(self.jdata["structures"], self.jdata["interaction"], [property])
        dfm_dirs_0 = glob.glob(os.path.join(target_path_0, "task.*"))
        for ii in dfm_dirs_0:
            self.assertTrue(os.path.isfile(os.path.join(ii, "STRU")))
            os.makedirs(os.path.join(ii, "OUT.ABACUS"))
            shutil.copy(
                os.path.join(ii, "STRU"), os.path.join(ii, "OUT.ABACUS", "STRU_ION_D")
            )

        new_prop_list = [
            {
                "type": "eos",
                "init_from_suffix": "00",
                "output_suffix": "02",
                "cal_setting": {
                    "relax_pos": True,
                    "relax_shape": True,
                    "relax_vol": False,
                },
            }
        ]
        make_property(
            self.jdata["structures"], self.jdata["interaction"], new_prop_list
        )
        self.assertTrue(os.path.isdir(path_to_work.replace("00", "02")))
        os.chdir(pwd)
        dfm_dirs_2 = glob.glob(os.path.join(target_path_2, "task.*"))
        self.assertEqual(len(dfm_dirs_2), len(dfm_dirs_0))



import argparse
import logging
from apex import (
    header,
    __version__,
)
from .run_step import run_step
from .submit import submit_workflow



def parse_args():
    parser = argparse.ArgumentParser(
        description=f"APEX: A scientific workflow for Alloy Properties EXplorer "
                    f"using simulations (v{__version__})\n"
                    f"Type 'apex -h' for help.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    subparsers = parser.add_subparsers(title="Valid subcommands", dest="cmd")
    # version
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"APEX v{__version__}"
    )
    ##########################################
    # Submit
    parser_submit = subparsers.add_parser(
        "submit",
        help="Submit an APEX workflow",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_submit.add_argument(
        "parameter", type=str, nargs='+',
        help='Json files to indicate calculation parameters'
    )
    parser_submit.add_argument(
        "-c", "--config",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config workflow",
    )

    parser_submit.add_argument(
        "-w", "--work",
        type=str, nargs='+',
        default='.',
        help="Working directory to be submitted",
    )

    parser_submit.add_argument(
        "-d", "--debug",
        action="store_true",
        help="Run APEX workflow via local debug mode"
    )
    parser_submit.add_argument(
        '-f', "--flow",
        choices=['relax', 'props', 'joint'],
        help="Specify type of workflow to submit: (relax | props | joint)"
    )

    ##########################################
    # Single step local test mode
    parser_test = subparsers.add_parser(
        "test",
        help="Single step local test mode",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser_test.add_argument(
        "parameter", type=str, nargs=1,
        help='Json file to indicate calculation parameters'
    )
    parser_test.add_argument(
        "step",
        type=str,
        choices=[
            'make_relax', 'run_relax', 'post_relax',
            'make_props', 'run_props', 'post_props'
        ],
        help="Specify step name to be tested: "
             "(make_relax | run_relax | post_relax |"
             " make_props | run_props | post_props)"
    )
    parser_test.add_argument(
        "-m", "--machine",
        type=str, nargs='?',
        default='./global.json',
        help="The json file to config the dpdispatcher",
    )

    parsed_args = parser.parse_args()
    # print help if no parser
    if not parsed_args.cmd:
        parser.print_help()

    return parser, parsed_args


def main():
    # logging
    logging.basicConfig(level=logging.INFO)
    # parse args
    parser, args = parse_args()
    header()
    if args.cmd == 'submit':
        submit_workflow(
            parameter=args.parameter,
            config_file=args.config,
            work_dir=args.work,
            flow_type=args.flow,
            is_debug=args.debug
        )

    elif args.cmd == 'test':
        run_step(
            parameter=args.parameter,
            machine_file=args.machine,
            step=args.step
        )
    else:
        raise RuntimeError(
            f"unknown command {args.command}\n{parser.print_help()}"
        )



import glob
import json
import logging
import os
import re

import numpy as np
from monty.serialization import dumpfn, loadfn
from pymatgen.analysis.defects.generators import VacancyGenerator
from pymatgen.core.structure import Structure

import apex.core.calculator.lib.abacus as abacus
from apex.core.Property import Property
from apex.core.refine import make_refine
from apex.core.reproduce import make_repro, post_repro
from dflow.python import upload_packages
upload_packages.append(__file__)


class Vacancy(Property):
    def __init__(self, parameter, inter_param=None):
        parameter["reproduce"] = parameter.get("reproduce", False)
        self.reprod = parameter["reproduce"]
        if not self.reprod:
            if not ("init_from_suffix" in parameter and "output_suffix" in parameter):
                default_supercell = [1, 1, 1]
                parameter["supercell"] = parameter.get("supercell", default_supercell)
                self.supercell = parameter["supercell"]
            parameter["cal_type"] = parameter.get("cal_type", "relaxation")
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": True,
                "relax_shape": True,
                "relax_vol": True,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
        else:
            parameter["cal_type"] = "static"
            self.cal_type = parameter["cal_type"]
            default_cal_setting = {
                "relax_pos": False,
                "relax_shape": False,
                "relax_vol": False,
            }
            if "cal_setting" not in parameter:
                parameter["cal_setting"] = default_cal_setting
            else:
                if "relax_pos" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_pos"] = default_cal_setting[
                        "relax_pos"
                    ]
                if "relax_shape" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_shape"] = default_cal_setting[
                        "relax_shape"
                    ]
                if "relax_vol" not in parameter["cal_setting"]:
                    parameter["cal_setting"]["relax_vol"] = default_cal_setting[
                        "relax_vol"
                    ]
            self.cal_setting = parameter["cal_setting"]
            parameter["init_from_suffix"] = parameter.get("init_from_suffix", "00")
            self.init_from_suffix = parameter["init_from_suffix"]
        self.parameter = parameter
        self.inter_param = inter_param if inter_param != None else {"type": "vasp"}

    def make_confs(self, path_to_work, path_to_equi, refine=False):
        path_to_work = os.path.abspath(path_to_work)
        if os.path.exists(path_to_work):
            #dlog.warning("%s already exists" % path_to_work)
            logging.warning("%s already exists" % path_to_work)
        else:
            os.makedirs(path_to_work)
        path_to_equi = os.path.abspath(path_to_equi)

        if "start_confs_path" in self.parameter and os.path.exists(
            self.parameter["start_confs_path"]
        ):
            init_path_list = glob.glob(
                os.path.join(self.parameter["start_confs_path"], "*")
            )
            struct_init_name_list = []
            for ii in init_path_list:
                struct_init_name_list.append(ii.split("/")[-1])
            struct_output_name = path_to_work.split("/")[-2]
            assert struct_output_name in struct_init_name_list
            path_to_equi = os.path.abspath(
                os.path.join(
                    self.parameter["start_confs_path"],
                    struct_output_name,
                    "relaxation",
                    "relax_task",
                )
            )

        task_list = []
        cwd = os.getcwd()

        if self.reprod:
            print("vacancy reproduce starts")
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            task_list = make_repro(
                self.inter_param,
                init_data_path,
                self.init_from_suffix,
                path_to_work,
                self.parameter.get("reprod_last_frame", False),
            )
            os.chdir(cwd)

        else:
            if refine:
                print("vacancy refine starts")
                task_list = make_refine(
                    self.parameter["init_from_suffix"],
                    self.parameter["output_suffix"],
                    path_to_work,
                )

                init_from_path = re.sub(
                    self.parameter["output_suffix"][::-1],
                    self.parameter["init_from_suffix"][::-1],
                    path_to_work[::-1],
                    count=1,
                )[::-1]
                task_list_basename = list(map(os.path.basename, task_list))

                for ii in task_list_basename:
                    init_from_task = os.path.join(init_from_path, ii)
                    output_task = os.path.join(path_to_work, ii)
                    os.chdir(output_task)
                    if os.path.isfile("supercell.json"):
                        os.remove("supercell.json")
                    if os.path.islink("supercell.json"):
                        os.remove("supercell.json")
                    os.symlink(
                        os.path.relpath(os.path.join(init_from_task, "supercell.json")),
                        "supercell.json",
                    )
                os.chdir(cwd)
            else:
                if self.inter_param["type"] == "abacus":
                    CONTCAR = abacus.final_stru(path_to_equi)
                    POSCAR = "STRU"
                else:
                    CONTCAR = "CONTCAR"
                    POSCAR = "POSCAR"

                equi_contcar = os.path.join(path_to_equi, CONTCAR)
                if not os.path.exists(equi_contcar):
                    raise RuntimeError("please do relaxation first")

                if self.inter_param["type"] == "abacus":
                    ss = abacus.stru2Structure(equi_contcar)
                else:
                    ss = Structure.from_file(equi_contcar)

                pre_vds = VacancyGenerator()
                vds = pre_vds.generate(ss)
                dss = []
                for jj in vds:
                    dss.append(
                        jj.get_supercell_structure(sc_mat=np.diag(self.supercell, k=0))
                    )

                print("gen vacancy with supercell " + str(self.supercell))
                os.chdir(path_to_work)
                if os.path.isfile(POSCAR):
                    os.remove(POSCAR)
                if os.path.islink(POSCAR):
                    os.remove(POSCAR)
                os.symlink(os.path.relpath(equi_contcar), POSCAR)
                #           task_poscar = os.path.join(output, 'POSCAR')

                for ii in range(len(dss)):
                    output_task = os.path.join(path_to_work, "task.%06d" % ii)
                    os.makedirs(output_task, exist_ok=True)
                    os.chdir(output_task)
                    for jj in [
                        "INCAR",
                        "POTCAR",
                        "POSCAR",
                        "conf.lmp",
                        "in.lammps",
                        "STRU",
                    ]:
                        if os.path.exists(jj):
                            os.remove(jj)
                    task_list.append(output_task)
                    dss[ii].to("POSCAR", "POSCAR")
                    if self.inter_param["type"] == "abacus":
                        abacus.poscar2stru("POSCAR", self.inter_param, "STRU")
                        os.remove("POSCAR")
                    # np.savetxt('supercell.out', self.supercell, fmt='%d')
                    dumpfn(self.supercell, "supercell.json")
                os.chdir(cwd)
        return task_list

    def post_process(self, task_list):
        pass

    def task_type(self):
        return self.parameter["type"]

    def task_param(self):
        return self.parameter

    def _compute_lower(self, output_file, all_tasks, all_res):
        output_file = os.path.abspath(output_file)
        res_data = {}
        ptr_data = os.path.dirname(output_file) + "\n"

        if not self.reprod:
            ptr_data += "Structure: \tVac_E(eV)  E(eV) equi_E(eV)\n"
            idid = -1
            for ii in all_tasks:
                idid += 1
                structure_dir = os.path.basename(ii)
                task_result = loadfn(all_res[idid])
                natoms = task_result["atom_numbs"][0]
                equi_path = os.path.abspath(
                    os.path.join(
                        os.path.dirname(output_file), "../relaxation/relax_task"
                    )
                )
                equi_result = loadfn(os.path.join(equi_path, "result.json"))
                equi_epa = equi_result["energies"][-1] / equi_result["atom_numbs"][0]
                evac = task_result["energies"][-1] - equi_epa * natoms

                supercell_index = loadfn(os.path.join(ii, "supercell.json"))
                ptr_data += "%s: %7.3f  %7.3f %7.3f \n" % (
                    str(supercell_index) + "-" + structure_dir,
                    evac,
                    task_result["energies"][-1],
                    equi_epa * natoms,
                )
                res_data[str(supercell_index) + "-" + structure_dir] = [
                    evac,
                    task_result["energies"][-1],
                    equi_epa * natoms,
                ]

        else:
            if "init_data_path" not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter["init_data_path"])
            res_data, ptr_data = post_repro(
                init_data_path,
                self.parameter["init_from_suffix"],
                all_tasks,
                ptr_data,
                self.parameter.get("reprod_last_frame", False),
            )

        with open(output_file, "w") as fp:
            json.dump(res_data, fp, indent=4)

        return res_data, ptr_data



import glob
import json
import os
import shutil
import sys
import unittest

import dpdata
import numpy as np

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
from pymatgen.io.vasp import Incar, Kpoints

from .context import make_kspacing_kpoints, setUpModule


class TestVASPMakeKpoint(unittest.TestCase):
    def test_gf_all(self):
        kspacing = 0.16
        gamma = False
        all_test = glob.glob(os.path.join("data.vasp.kp.gf", "test.*"))
        for ii in all_test:
            ret = make_kspacing_kpoints(os.path.join(ii, "POSCAR"), kspacing, gamma)
            kp = [int(jj) for jj in (ret.split("\n")[3].split())]
            kp_ref = list(np.loadtxt(os.path.join(ii, "kp.ref"), dtype=int))
            self.assertTrue(kp == kp_ref)



from abc import ABC, abstractmethod
from dflow.python import upload_packages
upload_packages.append(__file__)


class Task(ABC):
    @abstractmethod
    def __init__(self, inter_parameter, path_to_poscar):
        """
        Constructor

        Parameters
        ----------
        inter_parameter : dict
            A dict that specifies the interaction.
        path_to_poscar : str
            The path to POSCAR. Indicating in which system the task will be initialized.
        """
        pass

    @abstractmethod
    def make_potential_files(self, output_dir):
        """
        Prepare potential files for a computational task.
        For example, the VASP prepares POTCAR.
        DeePMD prepares frozen model(s).
        IMPORTANT: Interaction should be stored in output_dir/inter.json

        Parameters
        ----------
        output_dir : str
            The directory storing the potential files.

        Notes
        -----
        The following files are generated:

        inter.json: output file
                The task information is stored in `output_dir/inter.json`
        """
        pass

    @abstractmethod
    def make_input_file(self, output_dir, task_type, task_param):
        """
        Prepare input files for a computational task
        For example, the VASP prepares INCAR.
        LAMMPS (including DeePMD, MEAM...) prepares in.lammps.

        Parameters
        ----------
        output_dir : str
            The directory storing the input files.
        task_type : str
            Can be
            - "relaxation:": structure relaxation
            - "static": static computation calculates the energy, force... of a strcture
        task_param : dict
            The parameters of the task.
            For example the VASP interaction can be provided with
            { "ediff": 1e-6, "ediffg": 1e-5 }
        """
        pass

    @abstractmethod
    def compute(self, output_dir):
        """
        Compute output of the task.
        IMPORTANT: The output configuration should be converted and stored in a CONTCAR file.

        Parameters
        ----------
        output_dir : str
            The directory storing the input and output files.

        Returns
        -------
        result_dict: dict
            A dict that storing the result. For example:
            { "energy": xxx, "force": [xxx] }

        Notes
        -----
        The following files are generated:
        CONTCAR: output file
                The output configuration is converted to CONTCAR and stored in the `output_dir`
        """
        pass

    @property
    @staticmethod
    @abstractmethod
    def forward_files(self):
        """
        Return forward files.
        """
        pass

    @property
    @staticmethod
    @abstractmethod
    def forward_common_files(self):
        """
        Return forward common files.
        """
        pass

    @property
    @staticmethod
    @abstractmethod
    def backward_files(self):
        """
        Return backward files.
        """
        pass




import glob
import json
import os
import shutil
import sys
import unittest

import dpdata
import numpy as np
from monty.serialization import dumpfn, loadfn

from apex.core.calculator.Lammps import Lammps
from apex.core.calculator.lib.lammps import inter_deepmd

#from .context import make_kspacing_kpoints, setUpModule

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"


class TestLammps(unittest.TestCase):
    def setUp(self):
        self.jdata = {
            "structures": ["confs/std-fcc"],
            "interaction": {
                "type": "deepmd",
                "model": "lammps_input/frozen_model.pb",
                "deepmd_version": "1.1.0",
                "type_map": {"Al": 0},
            },
            "relaxation": {
                "cal_type": "relaxation",
                "cal_setting": {
                    "relax_pos": True,
                    "relax_shape": True,
                    "relax_vol": True,
                },
            },
        }

        self.equi_path = "confs/std-fcc/relaxation/relax_task"
        self.source_path = "equi/lammps"

        if not os.path.exists(self.equi_path):
            os.makedirs(self.equi_path)

        if not os.path.isfile(os.path.join(self.equi_path, "POSCAR")):
            shutil.copy(
                os.path.join(self.source_path, "Al-fcc.vasp"),
                os.path.join("confs/std-fcc", "POSCAR"),
            )

        self.confs = self.jdata["structures"]
        self.inter_param = self.jdata["interaction"]
        self.relax_param = self.jdata["relaxation"]
        self.Lammps = Lammps(
            self.inter_param, os.path.join(self.source_path, "Al-fcc.vasp")
        )

    def tearDown(self):
        if os.path.exists("confs/std-fcc/relaxation"):
            shutil.rmtree("confs/std-fcc/relaxation")

    def test_set_inter_type_func(self):
        self.Lammps.set_inter_type_func()
        self.assertEqual(inter_deepmd, self.Lammps.inter_func)

    def test_set_model_param(self):
        self.Lammps.set_model_param()
        model_param = {
            "model_name": ["frozen_model.pb"],
            "param_type": {"Al": 0},
            "deepmd_version": "1.1.0",
        }
        self.assertEqual(model_param, self.Lammps.model_param)

    def test_make_potential_files(self):
        cwd = os.getcwd()
        abs_equi_path = os.path.abspath(self.equi_path)
        self.Lammps.make_potential_files(abs_equi_path)
        self.assertTrue(os.path.islink(os.path.join(self.equi_path, "frozen_model.pb")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "inter.json")))
        ret = loadfn(os.path.join(self.equi_path, "inter.json"))
        self.assertEqual(self.inter_param, ret)
        os.chdir(cwd)

    def test_make_input_file(self):
        cwd = os.getcwd()
        abs_equi_path = os.path.abspath("confs/std-fcc/relaxation/relax_task")
        shutil.copy(
            os.path.join("confs/std-fcc", "POSCAR"),
            os.path.join(self.equi_path, "POSCAR"),
        )
        self.Lammps.make_input_file(abs_equi_path, "relaxation", self.relax_param)
        self.assertTrue(os.path.isfile(os.path.join(abs_equi_path, "conf.lmp")))
        self.assertTrue(os.path.islink(os.path.join(abs_equi_path, "in.lammps")))
        self.assertTrue(os.path.isfile(os.path.join(abs_equi_path, "task.json")))

    def test_forward_common_files(self):
        fc_files = ["in.lammps", "frozen_model.pb"]
        self.assertEqual(self.Lammps.forward_common_files(), fc_files)

    def test_backward_files(self):
        backward_files = ["log.lammps", "outlog", "dump.relax"]
        self.assertEqual(self.Lammps.backward_files(), backward_files)



import numpy as np

from dflow.python import upload_packages
upload_packages.append(__file__)

cartesian_ortho_norm_basis = np.array([ [1.0, 0.0, 0.0], \
                                        [0.0, 1.0, 0.0], \
                                        [0.0, 0.0, 1.0] ])

def plane_miller_bravais_to_miller(miller_bravais_plane):
    '''
    Convert a plane given in Miller-Bravais notation to Miller notation.
    '''

    # A plane in miller bravais notation is (h, k, i, l)
    h = miller_bravais_plane[0]
    k = miller_bravais_plane[1]
    i = miller_bravais_plane[2]
    l = miller_bravais_plane[3]

    # The plane converted to Miller notation (H, K, L)
    H = h
    K = k
    L = l

    return tuple(np.array([H, K, L], dtype='int_'))


def direction_miller_bravais_to_miller(miller_bravais_dir):
    '''
    Convert a direction given in Miller-Bravais notation to Miller notation.
    '''

    # The direction in Miller-Bravais notation is [U, V, T, W]
    U = miller_bravais_dir[0]
    V = miller_bravais_dir[1]
    T = miller_bravais_dir[2]
    W = miller_bravais_dir[3]

    # The direction converted to Miller notation [u, v, w]
    u = 2*U + V # U - t or 2*U + V
    v = 2*V + U # V - T or 2*V + U
    w = W

    return tuple(np.array([u, v, w], dtype='int_'))

def trans_mat_basis(dest, src=cartesian_ortho_norm_basis ):
    '''
    This matrix will transform any vector represented in basis
    A to a representation in basis B
    '''
    rmat = np.zeros((3, 3))

    rmat[0][0] = np.dot(dest[0], src[0])
    rmat[0][1] = np.dot(dest[0], src[1])
    rmat[0][2] = np.dot(dest[0], src[2])

    rmat[1][0] = np.dot(dest[1], src[0])
    rmat[1][1] = np.dot(dest[1], src[1])
    rmat[1][2] = np.dot(dest[1], src[2])

    rmat[2][0] = np.dot(dest[2], src[0])
    rmat[2][1] = np.dot(dest[2], src[1])
    rmat[2][2] = np.dot(dest[2], src[2])

    return rmat



import glob
import tempfile
from typing import Type
from multiprocessing import Pool
from dflow import config, s3_config
from dflow.python import OP
from monty.serialization import loadfn
import fpop, dpdata, apex
from apex.utils import get_task_type, get_flow_type
from .config import Config
from .flow import FlowFactory



def judge_flow(parameter, specify) -> (Type[OP], str, str, dict, dict):
    # identify type of flow and input parameter file
    num_args = len(parameter)
    if num_args == 1:
        task, run_op = get_task_type(loadfn(parameter[0]))
        flow = get_flow_type(loadfn(parameter[0]))
        task_type = task
        if flow == 'relax':
            flow_type = 'relax'
            if specify in ['props', 'joint']:
                raise RuntimeError(
                    'relaxation json file argument provided! Please check your jason file.'
                )
            relax_param = parameter[0]
            props_param = None
        elif flow == 'props':
            if specify in ['relax', 'joint']:
                raise RuntimeError(
                    'property test json file argument provided! Please check your jason file.'
                )
            flow_type = 'props'
            relax_param = None
            props_param = parameter[0]
        else:
            if specify == 'relax':
                flow_type = 'relax'
            elif specify == 'props':
                flow_type = 'props'
            else:
                flow_type = 'joint'
            relax_param = parameter[0]
            props_param = parameter[0]

    elif num_args == 2:
        task1, run_op1 = get_task_type(loadfn(parameter[0]))
        flow1 = get_flow_type(loadfn(parameter[0]))
        task2, run_op2 = get_task_type(loadfn(parameter[1]))
        flow2 = get_flow_type(loadfn(parameter[1]))
        if not flow1 == flow2:
            if specify == 'relax':
                flow_type = 'relax'
            elif specify == 'props':
                flow_type = 'props'
            else:
                flow_type = 'joint'
            if flow1 == 'relax' and flow2 == 'props':
                relax_param = parameter[0]
                props_param = parameter[1]
            elif flow1 == 'props' and flow2 == 'relax':
                relax_param = parameter[1]
                props_param = parameter[0]
            else:
                raise RuntimeError(
                    'confusion of jason arguments provided: '
                    'joint type of jason conflicts with the other json argument'
                )
        else:
            raise RuntimeError('Same type of input json files')
        if task1 == task2:
            task_type = task1
            run_op = run_op1
        else:
            raise RuntimeError('interaction types given are not matched')
    else:
        raise ValueError('A maximum of two input arguments is allowed')

    return run_op, task_type, flow_type, relax_param, props_param


def submit(flow,
           flow_type,
           work_dir,
           relax_param,
           props_param,
           conf=config,
           s3_conf=s3_config):
    # reset dflow global config for sub-processes
    config.update(conf)
    s3_config.update(s3_conf)

    if flow_type == 'relax':
        flow.submit_relax(
            work_dir=work_dir,
            relax_parameter=relax_param
        )
    elif flow_type == 'props':
        flow.submit_props(
            work_dir=work_dir,
            props_parameter=props_param
        )
    elif flow_type == 'joint':
        flow.submit_joint(
            work_dir=work_dir,
            props_parameter=props_param,
            relax_parameter=relax_param
        )


def submit_workflow(parameter,
                    config_file,
                    work_dir,
                    flow_type,
                    is_debug=False):
    try:
        config_dict = loadfn(config_file)
    except FileNotFoundError:
        raise FileNotFoundError(
            'Please prepare global.json under current work direction '
            'or use optional argument: -c to indicate a specific json file.'
        )
    # config dflow_config and s3_config
    wf_config = Config(**config_dict)
    wf_config.config_dflow(wf_config.dflow_config_dict)
    wf_config.config_bohrium(wf_config.bohrium_config_dict)
    wf_config.config_s3(wf_config.dflow_s3_config_dict)
    # set pre-defined dflow debug mode settings
    if is_debug:
        tmp_work_dir = tempfile.TemporaryDirectory()
        config["mode"] = "debug"
        config["debug_workdir"] = config_dict.get("debug_workdir", tmp_work_dir.name)
        s3_config["storage_client"] = None

    # judge basic flow info from user indicated parameter files
    (run_op, calculator, flow_type,
     relax_param, props_param) = judge_flow(parameter, flow_type)
    print(f'Running APEX calculation via {calculator}')
    print(f'Submitting {flow_type} workflow...')
    make_image = wf_config.basic_config_dict["apex_image_name"]
    run_image = wf_config.basic_config_dict[f"{calculator}_image_name"]
    if not run_image:
        run_image = wf_config.basic_config_dict["run_image_name"]
    run_command = wf_config.basic_config_dict[f"{calculator}_run_command"]
    if not run_command:
        run_command = wf_config.basic_config_dict["run_command"]
    post_image = make_image
    group_size = wf_config.basic_config_dict["group_size"]
    pool_size = wf_config.basic_config_dict["pool_size"]
    executor = wf_config.get_executor(wf_config.dispatcher_config_dict)
    upload_python_packages = wf_config.basic_config_dict["upload_python_packages"]
    upload_python_packages.extend(list(apex.__path__))
    upload_python_packages.extend(list(dpdata.__path__))
    upload_python_packages.extend(list(fpop.__path__))

    flow = FlowFactory(
        make_image=make_image,
        run_image=run_image,
        post_image=post_image,
        run_command=run_command,
        calculator=calculator,
        run_op=run_op,
        group_size=group_size,
        pool_size=pool_size,
        executor=executor,
        upload_python_packages=upload_python_packages
    )
    # submit the workflows
    work_dir_list = []
    for ii in work_dir:
        glob_list = glob.glob(ii)
        work_dir_list.extend(glob_list)
    if len(work_dir_list) > 1:
        n_processes = len(work_dir_list)
        pool = Pool(processes=n_processes)
        print(f'submitting via {n_processes} processes...')
        for ii in work_dir_list:
            res = pool.apply_async(
                submit,
                (flow, flow_type, ii, relax_param, props_param, config, s3_config)
            )
        pool.close()
        pool.join()
    elif len(work_dir_list) == 1:
        submit(flow, flow_type, work_dir_list[0], relax_param, props_param)

    print('Completed!')



import glob
import json
import os
import shutil
import sys
import unittest

import numpy as np
from monty.serialization import loadfn
from pymatgen.io.vasp import Incar

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"

from apex.core.calculator.VASP import VASP
from apex.core.calculator.lib.vasp import incar_upper


class TestVASP(unittest.TestCase):
    def setUp(self):
        self.jdata = {
            "structures": ["confs/hp-*"],
            "interaction": {
                "type": "vasp",
                "incar": "vasp_input/INCAR",
                "potcar_prefix": ".",
                "potcars": {"Li": "vasp_input/POTCAR"},
            },
            "relaxation": {
                "cal_type": "relaxation",
                "cal_setting": {
                    "relax_pos": True,
                    "relax_shape": True,
                    "relax_vol": True,
                },
            },
        }

        self.conf_path = "confs/hp-Li"
        self.equi_path = "confs/hp-Li/relaxation/relax_task"
        self.source_path = "equi/vasp"
        if not os.path.exists(self.equi_path):
            os.makedirs(self.equi_path)

        self.confs = self.jdata["structures"]
        inter_param = self.jdata["interaction"]
        self.task_param = self.jdata["relaxation"]
        self.VASP = VASP(inter_param, os.path.join(self.conf_path, "POSCAR"))

    def tearDown(self):
        if os.path.exists("confs/hp-Li/relaxation"):
            shutil.rmtree("confs/hp-Li/relaxation")
        if os.path.exists("inter.json"):
            os.remove("inter.json")
        if os.path.exists("POTCAR"):
            os.remove("POTCAR")

    def test_make_potential_files(self):
        if not os.path.exists(os.path.join(self.equi_path, "POSCAR")):
            with self.assertRaises(FileNotFoundError):
                self.VASP.make_potential_files(self.equi_path)
        shutil.copy(
            os.path.join(self.conf_path, "POSCAR"),
            os.path.join(self.equi_path, "POSCAR"),
        )
        self.VASP.make_potential_files(self.equi_path)
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "POTCAR")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "inter.json")))

    def test_make_input_file_1(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": True,
            "relax_shape": True,
            "relax_vol": False,
        }
        self.VASP.make_input_file(self.equi_path, "relaxation", param)
        incar = incar_upper(Incar.from_file(os.path.join(self.equi_path, "INCAR")))
        self.assertEqual(incar["ISIF"], 4)

    def test_make_input_file_2(self):
        self.VASP.make_input_file(self.equi_path, "relaxation", self.task_param)
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "task.json")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "KPOINTS")))
        self.assertTrue(os.path.isfile(os.path.join(self.equi_path, "INCAR")))
        incar = incar_upper(Incar.from_file(os.path.join(self.equi_path, "INCAR")))
        self.assertEqual(incar["ISIF"], 3)

    def test_make_input_file_3(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": True,
            "relax_shape": False,
            "relax_vol": False,
        }
        self.VASP.make_input_file(self.equi_path, "relaxation", param)
        incar = incar_upper(Incar.from_file(os.path.join(self.equi_path, "INCAR")))
        self.assertEqual(incar["ISIF"], 2)

    def test_make_input_file_4(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": False,
            "relax_shape": True,
            "relax_vol": False,
        }
        self.VASP.make_input_file(self.equi_path, "relaxation", param)
        incar = incar_upper(Incar.from_file(os.path.join(self.equi_path, "INCAR")))
        self.assertEqual(incar["ISIF"], 5)

    def test_make_input_file_5(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": False,
            "relax_shape": True,
            "relax_vol": True,
        }
        self.VASP.make_input_file(self.equi_path, "relaxation", param)
        incar = incar_upper(Incar.from_file(os.path.join(self.equi_path, "INCAR")))
        self.assertEqual(incar["ISIF"], 6)

    def test_make_input_file_5(self):
        param = self.task_param.copy()
        param["cal_setting"] = {
            "relax_pos": False,
            "relax_shape": True,
            "relax_vol": True,
            "kspacing": 0.01,
        }
        self.VASP.make_input_file(self.equi_path, "relaxation", param)
        incar = incar_upper(Incar.from_file(os.path.join(self.equi_path, "INCAR")))
        self.assertEqual(incar["ISIF"], 6)
        self.assertEqual(incar["KSPACING"], 0.01)

    def test_compute(self):
        ret = self.VASP.compute(os.path.join(self.conf_path, "relaxation"))
        self.assertIsNone(ret)
        shutil.copy(
            os.path.join(self.source_path, "OUTCAR"),
            os.path.join(self.equi_path, "OUTCAR"),
        )
        ret = self.VASP.compute(self.equi_path)
        ret_ref = loadfn(os.path.join(self.source_path, "outcar.json"))

        def compare_dict(dict1, dict2):
            self.assertEqual(dict1.keys(), dict2.keys())
            for key in dict1:
                if key == "stress":
                    self.assertTrue((np.array(dict1[key]["data"]) == dict2[key]).all())
                elif type(dict1[key]) is dict:
                    compare_dict(dict1[key], dict2[key])
                else:
                    if type(dict1[key]) is np.ndarray:
                        self.assertTrue((dict1[key] == dict2[key]).all())
                    else:
                        self.assertTrue(dict1[key] == dict2[key])

        compare_dict(ret, ret_ref.as_dict())

    def test_backward_files(self):
        backward_files = ["OUTCAR", "outlog", "CONTCAR", "OSZICAR", "XDATCAR"]
        self.assertEqual(self.VASP.backward_files(), backward_files)



#!/usr/bin/env python3

import argparse
import os
import re

from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.ext.matproj import Composition, MPRester

import apex.core.lib.crys as crys
from dflow.python import upload_packages
upload_packages.append(__file__)

global_std_crystal = {
    "fcc": crys.fcc,
    "hcp": crys.hcp,
    "dhcp": crys.dhcp,
    "bcc": crys.bcc,
    "diamond": crys.diamond,
    "sc": crys.sc,
}


def test_fit(struct, data):
    m = StructureMatcher()
    for ii in data:
        if m.fit(ii["structure"], struct):
            return True
    return False


def make_path_mp(ii):
    pf = ii["pretty_formula"]
    pf = re.sub("\d+", "", pf)
    task_id = ii["task_id"]
    work_path = "confs"
    work_path = os.path.join(work_path, pf)
    work_path = os.path.join(work_path, task_id)
    return work_path


def gen_ele_std(ele_name, ctype):
    struct = global_std_crystal[ctype](ele_name)
    work_path = "confs"
    work_path = os.path.join(work_path, ele_name)
    work_path = os.path.join(work_path, "std-" + ctype)
    os.makedirs(work_path, exist_ok=True)
    fposcar = os.path.join(work_path, "POSCAR")
    fjson = os.path.join(work_path, "data.json")
    struct.to("poscar", fposcar)
    return struct


def gen_element(ele_name, key):
    assert type(ele_name) == str
    mpr = MPRester(key)
    data = mpr.query(
        {"elements": [ele_name], "nelements": 1},
        properties=[
            "task_id",
            "pretty_formula",
            "formula",
            "anonymous_formula",
            "formation_energy_per_atom",
            "energy_per_atom",
            "structure",
        ],
    )
    for ii in data:
        work_path = make_path_mp(ii)
        os.makedirs(work_path, exist_ok=True)
        fposcar = os.path.join(work_path, "POSCAR")
        fjson = os.path.join(work_path, "data.json")
        ii["structure"].to("poscar", fposcar)
        ii["structure"].to("json", fjson)

    m = StructureMatcher()
    for ii in global_std_crystal.keys():
        ss = gen_ele_std(ele_name, ii)
        find = False
        for jj in data:
            if m.fit(ss, jj["structure"]):
                find = True
                break
        if find:
            work_path = make_path_mp(jj)
            with open(os.path.join(work_path, "std-crys"), "w") as fp:
                fp.write(ii + "\n")


def gen_element_std(ele_name):
    assert type(ele_name) == str
    for ii in global_std_crystal.keys():
        ss = gen_ele_std(ele_name, ii)


def gen_alloy(eles, key):

    mpr = MPRester(key)

    data = mpr.query(
        {"elements": {"$all": eles}, "nelements": len(eles)},
        properties=[
            "task_id",
            "pretty_formula",
            "formula",
            "anonymous_formula",
            "formation_energy_per_atom",
            "energy_per_atom",
            "structure",
        ],
    )
    if len(data) == 0:
        return

    alloy_file = make_path_mp(data[0])
    os.makedirs(alloy_file, exist_ok=True)
    alloy_file = os.path.join(alloy_file, "..")
    alloy_file = os.path.join(alloy_file, "alloy")
    with open(alloy_file, "w") as fp:
        None

    for ii in data:
        work_path = make_path_mp(ii)
        os.makedirs(work_path, exist_ok=True)
        fposcar = os.path.join(work_path, "POSCAR")
        fjson = os.path.join(work_path, "data.json")
        ii["structure"].to("poscar", fposcar)
        ii["structure"].to("json", fjson)


def _main():
    parser = argparse.ArgumentParser(description="gen structures")
    parser.add_argument("key", type=str, help="key id of material project")
    parser.add_argument(
        "elements", type=str, nargs="+", help="the list of appeared elements"
    )
    args = parser.parse_args()

    print("generate %s" % (args.elements))
    if len(args.elements) == 1:
        gen_element(args.elements[0], key)
        # gen_element_std(args.elements[0])
    else:
        gen_alloy(args.elements, key)


if __name__ == "__main__":
    _main()



import glob
import os
import re

import apex.core.calculator.lib.abacus as abacus
from dflow.python import upload_packages
upload_packages.append(__file__)


def make_refine(init_from_suffix, output_suffix, path_to_work):
    cwd = os.getcwd()
    init_from = re.sub(
        output_suffix[::-1], init_from_suffix[::-1], path_to_work[::-1], count=1
    )[::-1]
    if not os.path.exists(init_from):
        raise FileNotFoundError("the initial directory does not exist for refine")

    output = path_to_work
    init_from_task_tot = glob.glob(os.path.join(init_from, "task.[0-9]*[0-9]"))

    task_num = len(init_from_task_tot)

    task_list = []
    for ii in range(task_num):
        output_task = os.path.join(output, "task.%06d" % ii)
        os.makedirs(output_task, exist_ok=True)
        os.chdir(output_task)
        for jj in [
            "INCAR",
            "POTCAR",
            "POSCAR.orig",
            "POSCAR",
            "conf.lmp",
            "in.lammps",
            "STRU",
        ]:
            if os.path.exists(jj):
                os.remove(jj)
        task_list.append(output_task)
        init_from_task = os.path.join(init_from, "task.%06d" % ii)
        if not os.path.exists(init_from_task):
            raise FileNotFoundError(
                "the initial task directory does not exist for refine"
            )

        if os.path.isfile(os.path.join(init_from_task, "INPUT")) and os.path.isfile(
            os.path.join(init_from_task, "STRU")
        ):
            # if there has INPUT and STRU files in this path, we believe this is a ABACUS job
            CONTCAR = abacus.final_stru(init_from_task)
            POSCAR = "STRU"
        else:
            CONTCAR = "CONTCAR"
            POSCAR = "POSCAR"
        contcar = os.path.join(init_from_task, CONTCAR)
        init_poscar = os.path.join(init_from_task, POSCAR)
        if os.path.exists(contcar):
            os.symlink(os.path.relpath(contcar), POSCAR)
        elif os.path.exists(init_poscar):
            os.symlink(os.path.relpath(init_poscar), POSCAR)
        else:
            raise FileNotFoundError(
                "no %s or %s in the init_from directory" % (CONTCAR, POSCAR)
            )
    os.chdir(cwd)

    return task_list
